"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import os
import gemmi
import numpy
import argparse
import json
from servalcat.utils import logger
from servalcat import utils

def add_arguments(parser):
    parser.description = 'Shift back for Refmac local refinement results'
    parser.add_argument('--model',
                        help='Atomic model file that needs shift back')
    parser.add_argument('--refine_mtz',
                        help='Local-refined Refmac mtz file that needs shift back')
    parser.add_argument('--shifts',
                        required=True,
                        default="shifts.json",
                        help='Shift information file')
    parser.add_argument('--output_prefix',
                        help='output file prefix')
# add_arguments()

def parse_args(arg_list):
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    return parser.parse_args(arg_list)
# parse_args()

def refmac_mtz_in_original_cell(org_cell, org_grid_size, new_grid_size, shifts, mtz_in, mtz_out):
    targets = (("FWT", "PHWT"), ("DELFWT", "PHDELWT"))
    
    shifts_frac = org_cell.fractionalize(gemmi.Position(*shifts)).tolist()

    # Output mtz
    mtz = gemmi.Mtz()
    mtz.spacegroup = gemmi.SpaceGroup("P1")
    mtz.cell = org_cell
    mtz.add_dataset('HKL_base')
    for l in ["H", "K", "L"]: mtz.add_column(l, "H")

    data = None
    for i in range(len(targets)):
        d_min, m = utils.fileio.read_map_from_mtz(mtz_in, targets[i], new_grid_size)
        F = numpy.fft.fftn(m, org_grid_size).conj()
        grid = gemmi.ReciprocalComplexGrid(F.astype(numpy.complex64), mtz.cell, mtz.spacegroup)
        asu = grid.prepare_asu_data(dmin=d_min)
        if data is None:
            data = numpy.empty((len(asu), 3+2*len(targets)))
            data[:,:3] = asu.miller_array

        shift_factor = numpy.exp(-2j*numpy.pi*numpy.dot(asu.miller_array, shifts_frac))
        F_shift = asu.value_array * shift_factor
        data[:,3+2*i] = numpy.absolute(F_shift)
        data[:,3+2*i+1] = numpy.angle(F_shift, deg=True)

        mtz.add_column(targets[i][0], "F")
        mtz.add_column(targets[i][1], "P")
        
    mtz.set_data(data)
    mtz.write_to_file(mtz_out)
# refmac_mtz_in_original_cell()

def shift_back_model(st, shifts):
    # shifts must be Vec3 object
    
    for model in st:
        for cra in model.all():
            cra.atom.pos -= shifts

    if len(st.ncs) > 0:
        for n in st.ncs:
            newv = n.tr.vec - shifts + n.tr.mat.multiply(shifts)
            n.tr.vec.fromlist(newv.tolist())
# shift_back_model()

def shift_back(xyz_in, shifts_json, refine_mtz=None, out_prefix=None):
    logger.write("Reading shifts info from {}".format(shifts_json))
    info = json.load(open(shifts_json))
    for k in info:
        logger.write(" {}= {}".format(k, info[k]))

    org_cell = gemmi.UnitCell(*info["cell"])
    shifts = gemmi.Position(*info["shifts"])

    if refine_mtz:
        logger.write("Transforming MTZ: {}".format(refine_mtz))
        if out_prefix:
            mtz_out = out_prefix+".mtz"
        else:
            mtz_out = utils.fileio.splitext(os.path.basename(refine_mtz))[0] + "_shiftback.mtz"
            
        refmac_mtz_in_original_cell(org_cell,
                                    info["grid"],
                                    info["new_grid"],
                                    info["shifts"],
                                    refine_mtz,
                                    mtz_out)

    if xyz_in:
        logger.write("Shifting back model: {}".format(xyz_in))
        st, cif_ref = utils.fileio.read_structure_from_pdb_and_mmcif(xyz_in)

        st.cell = org_cell
        shift_back_model(st, shifts)
        prefix = out_prefix if out_prefix else utils.fileio.splitext(os.path.basename(xyz_in))[0] + "_shiftback"
        utils.fileio.write_model(st, prefix,
                                 pdb=True, cif=True, cif_ref=cif_ref)
        

# shift_back()


def main(args):
    if not args.model and not args.refine_mtz:
        logger.write("ERROR: give --model and/or --refine_mtz")
        return
    
    shift_back(args.model, args.shifts, args.refine_mtz, args.output_prefix)

if __name__ == "__main__":
    import sys
    args = parse_args(sys.argv[1:])
    main(args)
    



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
from . import logger
from . import symmetry
from . import keywords
from . import fileio
from . import hkl
from . import model
from . import maps
from . import refmac
from . import restraints
from . import commands



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
from servalcat.utils import logger
from servalcat.utils import fileio
from servalcat.utils import symmetry
from servalcat.utils import model
from servalcat.utils import restraints
import os
import gemmi
import numpy

def add_arguments(p):
    subparsers = p.add_subparsers(dest="subcommand")
    
    parser = subparsers.add_parser("show", description = 'Show file info supported by the program')
    parser.add_argument('files', nargs='+')

    parser = subparsers.add_parser("symmodel", description="Add symmetry annotation to model")
    parser.add_argument('--model', required=True)
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--map', help="Take box size from the map")
    group.add_argument('--cell', type=float, nargs=6, help="Box size")
    parser.add_argument('--pg', required=True, help="Point group symbol")
    parser.add_argument('--twist', type=float, help="Helical twist (degree)")
    parser.add_argument('--rise', type=float, help="Helical rise (Angstrom)")
    parser.add_argument('--chains', nargs="*", action="append", help="Select chains to keep")
    parser.add_argument('--biomt', action="store_true", help="Add BIOMT also")
    parser.add_argument('-o', '--output_prfix')
    parser.add_argument('--pdb', action="store_true", help="Write a pdb file")
    parser.add_argument('--cif', action="store_true", help="Write a cif file")
    
    parser = subparsers.add_parser("h_add", description = 'Add hydrogen in riding position')
    parser.add_argument('model')
    parser.add_argument('--ligand', nargs="*", action="append")
    parser.add_argument("--monlib",
                        help="Monomer library path. Default: $CLIBD_MON")
    parser.add_argument('-o','--output')
    parser.add_argument("--pos", choices=["elec", "nucl"], default="elec")

    parser = subparsers.add_parser("merge_models", description = 'Merge multiple model files')
    parser.add_argument('models', nargs="+")
    parser.add_argument('-o','--output', required=True)
# add_arguments()

def parse_args(arg_list):
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    return parser.parse_args(arg_list)
# parse_args()

def symmodel(args):
    if args.chains: args.chains = sum(args.chains, [])
    model_format = fileio.check_model_format(args.model)

    if (args.twist, args.rise).count(None) == 1:
        logger.write("ERROR: give both helical paramters --twist and --rise")
        return

    is_helical = args.twist is not None

    st = fileio.read_structure(args.model)
    st.spacegroup_hm = "P 1"
    start_xyz = numpy.zeros(3)
    if args.map:
        logger.write("Reading cell from map")
        g, grid_start = fileio.read_ccp4_map(args.map)
        st.cell = g.unit_cell
        start_xyz = numpy.array(g.get_position(*grid_start).tolist())
    elif args.cell:
        st.cell = gemmi.UnitCell(*args.cell)
    elif not st.cell.is_crystal():
        logger.write("Error: Unit cell parameters look wrong. Please use --map or --cell")
        return

    if args.chains:
        logger.write("Keep {} chains only".format(" ".join(args.chains)))
        chains = set(args.chains)
        for m in st:
            to_del = [c.name for c in m if c.name not in chains]
            for c in to_del: m.remove_chain(c)

    all_chains = [c.name for c in st[0] if c.name not in st[0]]

    A = numpy.array(st.cell.orthogonalization_matrix.tolist())
    center = numpy.sum(A, axis=1) / 2 + start_xyz

    if is_helical:
        ncsops = symmetry.generate_helical_operators(st, start_xyz, center,
                                                  args.pg, args.twist, args.rise)
        logger.write("{} helical operators found".format(len(ncsops)))
    else:
        _, _, ops = symmetry.operators_from_symbol(args.pg)
        logger.write("{} operators found for {}".format(len(ops), args.pg))
        symmetry.show_operators_axis_angle(ops)
        ncsops = symmetry.make_NcsOps_from_matrices(ops, cell=st.cell, center=center)

    st.ncs.clear()
    st.ncs.extend([x for x in ncsops if not x.tr.is_identity()])

    if args.biomt:
        st.assemblies.clear()
        st.raw_remarks = []
        a = gemmi.Assembly("1")
        g = gemmi.Assembly.Gen()
        for i, nop in enumerate(ncsops):
            op = gemmi.Assembly.Operator()
            op.transform = nop.tr
            if not nop.tr.is_identity(): op.type = "point symmetry operation"
            g.operators.append(op)
        g.chains = all_chains
        a.generators.append(g)
        a.special_kind = gemmi.AssemblySpecialKind.CompletePoint
        st.assemblies.append(a)

    if not args.output_prfix:
        args.output_prfix = fileio.splitext(os.path.basename(args.model))[0] + "_asu"

    if args.pdb or args.cif:
        fileio.write_model(st, args.output_prfix, pdb=args.pdb, cif=args.cif)
    else:
        fileio.write_model(st, file_name=args.output_prfix+model_format)

    # Sym expand
    model.expand_ncs(st)
    st.assemblies.clear()
    args.output_prfix += "_expanded"
    if args.pdb or args.cif:
        fileio.write_model(st, args.output_prfix, pdb=args.pdb, cif=args.cif)
    else:
        fileio.write_model(st, file_name=args.output_prfix+model_format)
# symmodel()

def h_add(args):
    st = fileio.read_structure(args.model)
    resnames = st[0].get_all_residue_names()
    model_format = fileio.check_model_format(args.model)
    
    if not args.output:
        tmp = fileio.splitext(os.path.basename(args.model))[0]
        args.output = tmp + "_h" + model_format
        logger.write("Output file: {}".format(args.output))
        
    args.ligand = sum(args.ligand, []) if args.ligand else []
    monlib = restraints.load_monomer_library(resnames,
                                             monomer_dir=args.monlib,
                                             cif_files=args.ligand)
    restraints.add_hydrogens(st, monlib, args.pos)
    fileio.write_model(st, file_name=args.output)
# h_add()

def merge_models(args):
    logger.write("Reading file   1: {}".format(args.models[0]))
    st = fileio.read_structure(args.models[0])
    logger.write("                  chains {}".format(" ".join([c.name for c in st[0]])))

    for i, f in enumerate(args.models[1:]):
        logger.write("Reading file {:3d}: {}".format(i+2, f))
        st2 = fileio.read_structure(f)
        for c in st2[0]:
            org_id = c.name
            c2 = st[0].add_chain(c, unique_name=True)
            if c.name != c2.name:
                logger.write("                  chain {} merged (ID changed to {})".format(c.name, c2.name))
            else:
                logger.write("                  chain {} merged".format(c.name))

    fileio.write_model(st, file_name=args.output)
# merge_models()

def show(args):
    for filename in args.files:
        ext = fileio.splitext(filename)[1]
        if ext in (".mrc", ".ccp4", ".map"):
            fileio.read_ccp4_map(filename)
            logger.write("\n")
# show()

def main(args):
    com = args.subcommand
    if com == "show":
        show(args)
    elif com == "symmodel":
        symmodel(args)
    elif com == "h_add":
        h_add(args)
    elif com == "merge_models":
        merge_models(args)
# main()

if __name__ == "__main__":
    import sys
    args = parse_args(sys.argv[1:])
    main(args)



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import sys
import datetime
import traceback

class Logger(object):
    def __init__(self, file_out=None, append=True):
        self.ofs = None
        if file_out:
            self.set_file(file_out, append)
    # __init__()
    
    def set_file(self, file_out, append=True):
        self.ofs = open(file_out, "a" if append else "w")
    # set_file()

    def write(self, l, end="\n", flush=True):
        print(l, end=end)
        if self.ofs is not None:
            self.ofs.write(l)
            self.ofs.write(end)
            if flush: self.ofs.flush()
    # write()

    def close(self):
        self.ofs.close()
        self.ofs = None
    # close()
# class Logger

_logger = Logger() # singleton
set_file = _logger.set_file
write = _logger.write
close = _logger.close


def handle_exception(exc_type, exc_value, exc_traceback):
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return

    name = type(exc_value).__name__ if hasattr(type(exc_value), "__name__") else "(unknown)"
    #_logger.write("Uncaught exception: {}: {}".format(name, exc_value))
    _logger.write("".join(traceback.format_exception(exc_type, exc_value, exc_traceback)))
    _logger.write("# Abnormally finished on {}\n".format(datetime.datetime.now()))

# handle_exception()

sys.excepthook = handle_exception



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import numpy
from servalcat.utils import logger
from servalcat.utils import hkl

def mask_from_model():
    pass

def half2full(map_h1, map_h2):
    assert map_h1.shape == map_h2.shape
    assert map_h1.unit_cell == map_h2.unit_cell
    tmp = (numpy.array(map_h1)+numpy.array(map_h2))/2.
    gr = gemmi.FloatGrid(tmp, map_h1.unit_cell, map_h1.spacegroup)
    return gr
# half2full()

def sharpen_mask_unsharpen(maps, mask, d_min, b=None):
    assert len(maps) < 3
    if b is None and len(maps) != 2:
        raise RuntimeError("Cannot determine sharpening")

    hkldata = mask_and_fft_maps(maps, d_min)
    normalizer = numpy.ones(len(hkldata.df.index))
    if len(maps) == 2:
        labs = ["F_map1", "F_map2"]
    else:
        labs = ["FP"]
        
    # 1. Sharpen
    if b is None:
        hkldata.setup_relion_binning()
        calc_noise_var_from_halfmaps(hkldata)
        FP = hkldata.df.FP.to_numpy()
        logger.write("""$TABLE: Normalizing before masking:
$GRAPHS: ln(Mn(|F|)) :A:1,2:
: Normalizer :A:1,3:
: FSC(full) :A:1,4:
$$ 1/resol^2 ln(Mn(|F|)) normalizer FSC $$
$$""")
        for i_bin, bin_d_max, bin_d_min in hkldata.bin_and_limits():
            sel = i_bin == hkldata.df.bin
            Fo = FP[sel]
            FSCfull = hkldata.binned_df.FSCfull[i_bin]
            sig_fo = numpy.std(Fo)
            if FSCfull > 0:
                n_fo = sig_fo * numpy.sqrt(FSCfull)
            else:
                n_fo = sig_fo # XXX not a right way
                
            normalizer[sel] = n_fo
            for lab in labs: hkldata.df.loc[sel, lab] /= n_fo
            logger.write("{:.4f} {:.2f} {:.3f} {:.4f}".format(1/bin_d_min**2,
                                                              numpy.log(numpy.average(numpy.abs(Fo))),
                                                              n_fo, FSCfull))

        logger.write("$$")

    else:
        logger.write("Sharpening B before masking= {}".format(b))
        s2 = 1./hkldata.d_spacings()**2
        normalizer[:] = numpy.exp(-b*s2/4.)
        for lab in labs: hkldata.df.loc[:, lab] /= normalizer

    # 2. Mask
    new_maps = []
    for lab in labs:
        m = hkldata.fft_map(lab, grid_size=mask.shape)
        new_maps.append([gemmi.FloatGrid(numpy.array(m)*mask, hkldata.cell, hkldata.sg), None])

    # 3. Unsharpen
    hkldata = mask_and_fft_maps(new_maps, d_min)
    for lab in labs: hkldata.df[lab] *= normalizer
    
    new_maps = []
    for i, lab in enumerate(labs):
        m = hkldata.fft_map(lab, grid_size=mask.shape)
        new_maps.append([m]+maps[i][1:])

    return new_maps
# sharpen_mask_unsharpen()

def mask_and_fft_maps(maps, d_min, mask=None):
    assert len(maps) <= 2
    asus = []
    for m in maps:
        g = m[0]
        if mask is not None:
            g = gemmi.FloatGrid(numpy.array(g)*mask,
                                g.unit_cell, g.spacegroup)

        asus.append(gemmi.transform_map_to_f_phi(g).prepare_asu_data(dmin=d_min))

    if len(maps) == 2:
        df = hkl.df_from_asu_data(asus[0], "F_map1")
        hkldata = hkl.HklData(asus[0].unit_cell, asus[0].spacegroup, False, df)
        hkldata.merge_asu_data(asus[1], "F_map2")
        hkldata.df["FP"] = (hkldata.df.F_map1 + hkldata.df.F_map2)/2.
    else:
        df = hkl.df_from_asu_data(asus[0], "FP")
        hkldata = hkl.HklData(asus[0].unit_cell, asus[0].spacegroup, False, df)
        
    return hkldata
# mask_and_fft_maps()
    
def calc_noise_var_from_halfmaps(hkldata):
    # Scale
    #iniscale = utils.scaling.InitialScaler(fo_asu, fc_asu, aniso=True)
    #iniscale.run()
    #scale_for_fo = iniscale.get_scales()
    #fo_asu.value_array[:] *= scale_for_fo
    #asu1.value_array[:] *= scale_for_fo
    #asu2.value_array[:] *= scale_for_fo

    s_array = 1./hkldata.d_spacings()
    hkldata.binned_df["var_noise"] = 0.
    hkldata.binned_df["var_signal"] = 0.
    hkldata.binned_df["FSCfull"] = 0.
    
    logger.write("Bin Ncoeffs d_max   d_min   FSChalf var.noise   scale")
    F_map1 = hkldata.df.F_map1.to_numpy()
    F_map2 = hkldata.df.F_map2.to_numpy()
    for i_bin, bin_d_max, bin_d_min in hkldata.bin_and_limits():
        sel = i_bin == hkldata.df.bin
        # scale
        scale = 1. #numpy.sqrt(var_cmpl(fc)/var_cmpl(fo))
        #hkldata.df.loc[sel, "FP"] *= scale
        #hkldata.df.loc[sel, "F_map1"] *= scale
        #hkldata.df.loc[sel, "F_map2"] *= scale
        
        sel1 = F_map1[sel]
        sel2 = F_map2[sel]

        if sel1.size < 3:
            logger.write("WARNING: skipping bin {} with size= {}".format(i_bin, sel1.size))
            continue

        fsc = numpy.real(numpy.corrcoef(sel1, sel2)[1,0])
        varn = numpy.var(sel1-sel2)/4
        vart = numpy.var(sel1+sel2)/4
        logger.write("{:3d} {:7d} {:7.3f} {:7.3f} {:.4f} {:e} {}".format(i_bin, sel1.size, bin_d_max, bin_d_min,
                                                                         fsc, varn, scale))
        hkldata.binned_df.loc[i_bin, "var_noise"] = varn
        hkldata.binned_df.loc[i_bin, "var_signal"] = vart-varn
        hkldata.binned_df.loc[i_bin, "FSCfull"] = 2*fsc/(1+fsc)
# calc_noise_var_from_halfmaps()

def write_ccp4_map(filename, array, cell=None, sg=None, mask_for_extent=None, grid_start=None):
    if type(array) == numpy.ndarray:
        # TODO check dtype
        if sg is None: sg = gemmi.SpaceGroup(1)
        grid = gemmi.FloatGrid(array, cell, sg)
    else:
        # TODO check type
        grid = array
        
    ccp4 = gemmi.Ccp4Map()
    ccp4.grid = grid
    ccp4.update_ccp4_header(2, True) # float, update stats

    if mask_for_extent is not None:
        tmp = numpy.where(numpy.array(mask_for_extent)>0)
        if grid_start is not None:
            grid_start = numpy.array(grid_start)[:,None]
            grid_shape = numpy.array(grid.shape)[:,None]
            tmp -= grid_start
            tmp += (grid_shape*numpy.floor(1-tmp/grid_shape)).astype(int) + grid_start

        l = [(min(x), max(x)) for x in tmp]
        box = gemmi.FractionalBox()
        for i in (0, 1):
            fm = grid.get_fractional(l[0][i], l[1][i], l[2][i])
            box.extend(fm)

        logger.write(" setting extent: {} {}".format(box.minimum, box.maximum))
        ccp4.set_extent(box)
    elif grid_start is not None:
        logger.write(" setting starting grid: {} {} {}".format(*grid_start))
        new_grid = gemmi.FloatGrid(grid.get_subarray(*(list(grid_start)+list(grid.shape))),
                                   cell,
                                   sg)
        ccp4.grid = new_grid
        for i in range(3):
            ccp4.set_header_i32(5+i, grid_start[i])

    ccp4.write_ccp4_map(filename)
# write_ccp4_map()



# TODO shift map first using mask!
"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import numpy
import time
from servalcat.utils import logger
from servalcat import utils
import argparse

def add_arguments(parser):
    parser.description = 'Fo-Fc map calculation based on model and data errors'
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--halfmaps", nargs=2)
    group.add_argument("--map", help="Use only if you really do not have half maps.")
    parser.add_argument('--pixel_size', type=float,
                        help='Override pixel size (A)')
    parser.add_argument('--model', required=True,
                        help='Input atomic model file (PDB or mmCIF/PDBx')
    parser.add_argument("-d", '--resolution', type=float, required=True)
    parser.add_argument('-m', '--mask', help="mask file")
    parser.add_argument('-r', '--mask_radius', type=float, help="mask radius")
    parser.add_argument("-B", type=float, help="Estimated blurring.")
    parser.add_argument("--half1_only", action='store_true', help="Only use half 1 for map (use half 2 only for noise estimation)")
    parser.add_argument("--normalized_map", action='store_true',
                        help="Write normalized map in the masked region")
    parser.add_argument("--crop", action='store_true',
                        help="Write cropped maps")
    parser.add_argument("--monlib",
                        help="Monomer library path. Default: $CLIBD_MON")
    parser.add_argument("--omit_proton", action='store_true',
                        help="Omit proton from model in map calculation")
    parser.add_argument("--omit_h_electron", action='store_true',
                        help="Omit hydrogen electrons from model in map calculation")
    parser.add_argument('-o','--output_prefix', default="diffmap",
                        help='output file name prefix')
# add_arguments()

def parse_args(arg_list):
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    return parser.parse_args(arg_list)
# parse_args()

def calc_D_and_S(hkldata, has_halfmaps=True, half1_only=False):#fo_asu, fc_asu, varn, bins, bin_idxes):
    bdf = hkldata.binned_df
    bdf["D"] = 0.
    bdf["S"] = 0.
    stats_str = """$TABLE: Statistics :
$GRAPHS
: log(Mn(|F|^2)) and variances :A:1,6,7,8,13,14:
: FSC :A:1,9,10,11:
: weights :A:1,12,15,16:
$$
1/resol^2 bin n d_max d_min log(var(Fo)) log(var(Fc)) log(var(DFc)) FSC.model FSC.full sqrt(FSC.full) D log(var_U,T) log(var_noise) wFo wFc
$$
$$
"""
    tmpl = "{:.4f} {:3d} {:7d} {:7.3f} {:7.3f} {:.4e} {:.4e} {:4e} {: .4f}   {: .4f} {: .4f} {: .4e} {:.4e} {:.4e} {:.4f} {:.4f}\n"

    var_noise = None
    FP = hkldata.df.FP.to_numpy()
    if half1_only:
        FP = hkldata.df.F_map1.to_numpy()
        var_noise = hkldata.binned_df.var_noise * 2
    elif has_halfmaps:
        var_noise = hkldata.binned_df.var_noise
        
    FC = hkldata.df.FC.to_numpy()
    for i_bin, bin_d_max, bin_d_min in hkldata.bin_and_limits():
        sel = i_bin == hkldata.df.bin
        Fo = FP[sel]
        Fc = FC[sel]
        fsc = numpy.real(numpy.corrcoef(Fo, Fc)[1,0])
        bdf.loc[i_bin, "D"] = numpy.sum(numpy.real(Fo * numpy.conj(Fc)))/numpy.sum(numpy.abs(Fc)**2)
        if has_halfmaps:
            varn = var_noise[i_bin]
            fsc_full = hkldata.binned_df.FSCfull[i_bin]
            S = max(0, numpy.average(numpy.abs(Fo-bdf.D[i_bin]*Fc)**2)-varn)
            bdf.loc[i_bin, "S"] = S
            w = S/(S+varn)
        else:
            varn = fsc_full = 0
            w = 1

        with numpy.errstate(divide="ignore"):
            stats_str += tmpl.format(1/bin_d_min**2, i_bin, Fo.size, bin_d_max, bin_d_min,
                                     numpy.log(numpy.average(numpy.abs(Fo)**2)),
                                     numpy.log(numpy.average(numpy.abs(Fc)**2)),
                                     numpy.log(bdf.D[i_bin]**2*numpy.average(numpy.abs(Fc)**2)),
                                     fsc, fsc_full, numpy.sqrt(fsc_full), bdf.D[i_bin],
                                     numpy.log(bdf.S[i_bin]), numpy.log(varn),
                                     w, 1-w)
    return stats_str
# calc_D_and_S()

#import line_profiler
#profile = line_profiler.LineProfiler()
#import atexit
#atexit.register(profile.print_stats)
#@profile
def calc_maps(hkldata, B=None, has_halfmaps=True, half1_only=False):
    if has_halfmaps:
        labs = ["Fupdate", "DELFWT", "FWT", "DELFWT_noscale", "Fupdate_noscale"]
        if B is not None: labs.extend(["Fupdate_b0", "DELFWT_b0", "FWT_b0"])
    else:
        labs = ["DELFWT"]
        
    tmp = {}
    for l in labs:
        tmp[l] = numpy.zeros(len(hkldata.df.index), numpy.complex128)

    logger.write("Calculating maps..")
    time_t = time.time()

    if half1_only:
        FP = hkldata.df.F_map1.to_numpy()
    else:
        FP = hkldata.df.FP.to_numpy()
        
    FC = hkldata.df.FC.to_numpy()
    
    for i_bin, bin_d_max, bin_d_min in hkldata.bin_and_limits():
        sel = i_bin == hkldata.df.bin
        Fo = FP[sel]
        Fc = FC[sel]
        D = hkldata.binned_df.D[i_bin]
        if not has_halfmaps:
            delfwt = (Fo-D*Fc)
            tmp["DELFWT"][sel] = delfwt
            continue

        S = hkldata.binned_df.S[i_bin]
        fsc = hkldata.binned_df.FSCfull[i_bin]
        if half1_only:
            varn = hkldata.binned_df.var_noise[i_bin] * 2
            fsc = fsc/(2-fsc) # to FSChalf
        else:
            varn = hkldata.binned_df.var_noise[i_bin]

        delfwt = (Fo-D*Fc)*S/(S+varn)
        fup = (Fo*S+varn*D*Fc)/(S+varn)

        tmp["DELFWT_noscale"][sel] = delfwt
        tmp["Fupdate_noscale"][sel] = fup

        sig_fo = numpy.std(Fo)
        n_fo = sig_fo * numpy.sqrt(fsc)
        if n_fo < 1e-10 or n_fo != n_fo:
            logger.write("WARNING: skipping bin {} sig_fo={} fsc={}".format(i_bin, sig_fo, fsc))
            continue
        #n_fofc = numpy.sqrt(var_cmpl(Fo-D*Fc))

        lab_suf = "" if B is None else "_b0"
        tmp["FWT"+lab_suf][sel] = numpy.sqrt(fsc)*Fo/sig_fo
        tmp["DELFWT"+lab_suf][sel] = delfwt/n_fo
        tmp["Fupdate"+lab_suf][sel] = fup/n_fo

        if B is not None:
            s2 = 1./hkldata.d_spacings()[sel]**2
            k = numpy.exp(-B*s2/4.)
            k2 = numpy.exp(-B*s2/2.)
            fsc_l = k2*fsc/(1+(k2-1)*fsc)
            #n_fo = sig_fo * numpy.sqrt(fsc_l) * k
            S_l = S * k2
            
            delfwt = (Fo-D*Fc)*S*k/(S_l+varn)/n_fo # S_l/k = S*k
            fup = (Fo*S_l+varn*D*Fc)/(S_l+varn)/n_fo/k
            fwt = Fo*fsc_l/n_fo
            logger.write("{:4d} {:.4e} {:.4e} {:.4e} {:.4e} {:.4e} {:.4e}".format(i_bin,
                                                                           numpy.average(n_fo),
                                                                           numpy.average(sig_fo),
                                                                           numpy.average(fsc_l),
                                                                           numpy.average(k),
                                                                           numpy.average(abs(fup)),
                                                                           numpy.average(abs(delfwt))))
            tmp["FWT"][sel] = fwt
            tmp["DELFWT"][sel] = delfwt
            tmp["Fupdate"][sel] = fup

    for l in labs:
        hkldata.df[l] = tmp[l]

    logger.write(" finished in {:.3f} sec.".format(time.time()-time_t))
    return labs
# calc_maps()

def dump_to_mtz(hkldata, map_labs, mtz_out):
    map_labs = map_labs + ["FP", "FC"]
    data = numpy.empty((len(hkldata.df.index), len(map_labs)*2+3))
    data[:,:3] = hkldata.df[["H","K","L"]]
    for i, lab in enumerate(map_labs):
        data[:,3+i*2] = numpy.abs(hkldata.df[lab])
        data[:,3+i*2+1] = numpy.angle(hkldata.df[lab], deg=True)
        
    mtz = gemmi.Mtz()
    mtz.spacegroup = hkldata.sg
    mtz.cell = hkldata.cell
    mtz.add_dataset('HKL_base')
    for label in ['H', 'K', 'L']: mtz.add_column(label, 'H')

    for lab in map_labs:
        mtz.add_column(lab, "F")
        mtz.add_column(("PH"+lab).replace("FWT", "WT"), "P")
    
    mtz.set_data(data)
    mtz.write_to_file(mtz_out)
# dump_to_mtz()

def calc_fofc(st, d_min, maps, mask=None, monlib=None, B=None, half1_only=False, omit_proton=False, omit_h_electron=False):
    fc_asu = utils.model.calc_fc_fft(st, d_min, cutoff=1e-7, monlib=monlib, source="electron")
    hkldata = utils.maps.mask_and_fft_maps(maps, d_min, mask)
    hkldata.merge_asu_data(fc_asu, "FC")
    hkldata.setup_relion_binning()

    has_halfmaps = (len(maps) == 2)
    if has_halfmaps:
        utils.maps.calc_noise_var_from_halfmaps(hkldata)

    if B is not None:
        Bave = numpy.average(utils.model.all_B(st))
        logger.write("Using user-specified B: {}".format(B))
        logger.write("    Average B of model= {:.2f}".format(Bave))
        b_local = B - Bave
        logger.write("    Relative B for map= {:.2f}".format(b_local))
    else:
        b_local = None
        
    stats_str = calc_D_and_S(hkldata, has_halfmaps=has_halfmaps, half1_only=half1_only)

    if omit_proton or omit_h_electron:
        fc_asu_2 = utils.model.calc_fc_fft(st, d_min, cutoff=1e-7, monlib=monlib, source="electron",
                                           omit_proton=omit_proton, omit_h_electron=omit_h_electron)
        del hkldata.df["FC"]
        hkldata.merge_asu_data(fc_asu_2, "FC")
    
    map_labs = calc_maps(hkldata, B=b_local, has_halfmaps=has_halfmaps, half1_only=half1_only)
    return hkldata, map_labs, stats_str
# calc_fofc()

def write_files(hkldata, map_labs, grid_start, stats_str,
                mask=None, output_prefix="diffmap", crop=False, normalize_map=False, omit_h_electron=False):
    dump_to_mtz(hkldata, map_labs, "{}.mtz".format(output_prefix))
    open("{}_Fstats.log".format(output_prefix), "w").write(stats_str)
    
    # DEBUG
    if mask is not None:
        utils.maps.write_ccp4_map("fc_map.mrc",  hkldata.fft_map("FC", grid_size=mask.shape), cell=hkldata.cell,
                                  mask_for_extent=mask if crop else None,
                                  grid_start=grid_start)
        hkldata.df["DFC"] = hkldata.df["FC"] * hkldata.binned_data_as_array("D")
        utils.maps.write_ccp4_map("dfc_map.mrc",  hkldata.fft_map("DFC", grid_size=mask.shape), cell=hkldata.cell,
                                  mask_for_extent=mask if crop else None,
                                  grid_start=grid_start)

    
    if normalize_map and mask is not None:
        logger.write("Normalized Fo-Fc map requested.")
        delfwt_map = hkldata.fft_map("DELFWT", grid_size=mask.shape)
        cutoff = 0.5
        masked = numpy.array(delfwt_map)[mask>cutoff]
        logger.write("   Whole volume: {} voxels".format(delfwt_map.point_count))
        logger.write("  Masked volume: {} voxels (>{})".format(masked.size, cutoff))
        global_mean = numpy.average(delfwt_map)
        global_std = numpy.std(delfwt_map)
        logger.write("    Global mean: {:.3e}".format(global_mean))
        logger.write("     Global std: {:.3e}".format(global_std))
        masked_mean = numpy.average(masked)
        masked_std = numpy.std(masked)
        logger.write("    Masked mean: {:.3e}".format(masked_mean))
        logger.write("     Masked std: {:.3e}".format(masked_std))
        #logger.write(" If you want to scale manually: {}".format())
        scaled = (delfwt_map - masked_mean)/masked_std
        if omit_h_electron:
            scaled *= -1
            filename = "{}_normalized_fofc_flipsign.mrc".format(output_prefix)
        else:
            filename = "{}_normalized_fofc.mrc".format(output_prefix)
        logger.write("  Writing {}".format(filename))
        utils.maps.write_ccp4_map(filename, scaled, cell=hkldata.cell,
                                  mask_for_extent=mask if crop else None,
                                  grid_start=grid_start)

        # Write Fo map as well
        fwt_map = hkldata.fft_map("FWT", grid_size=mask.shape)
        masked = numpy.array(fwt_map)[mask>cutoff]
        masked_mean = numpy.average(masked)
        masked_std = numpy.std(masked)
        scaled = (fwt_map - masked_mean)/masked_std # does not make much sense for Fo map though
        filename = "{}_normalized_fo.mrc".format(output_prefix)
        logger.write("  Writing {}".format(filename))
        utils.maps.write_ccp4_map(filename, scaled, cell=hkldata.cell,
                                  mask_for_extent=mask if crop else None,
                                  grid_start=grid_start)

# write_files()

def main(args):
    if not args.halfmaps and not args.map:
        logger.write("Error: give --halfmaps or --map")
        return

    if not args.halfmaps and args.B is not None:
        logger.write("Error: -B only works for half maps")
        return

    if args.half1_only:
        if not args.halfmaps:
            logger.write("--half1_only requires half maps")
            return
        logger.write("--half1_only specified. Half map 2 is used only for noise estimation")

    if not args.halfmaps:
        logger.write("Warning: using --halfmaps is strongly recommended!")

    st = utils.fileio.read_structure(args.model)
    utils.model.expand_ncs(st)

    if (args.omit_proton or args.omit_h_electron) and st[0].count_hydrogen_sites() == 0:
        logger.write("ERROR! --omit_proton/--omit_h_electron requested, but no hydrogen atoms were found.")
        return

    if args.halfmaps:
        maps = [utils.fileio.read_ccp4_map(f, pixel_size=args.pixel_size) for f in args.halfmaps]
        has_halfmaps = True
    else:
        maps = [utils.fileio.read_ccp4_map(args.map, pixel_size=args.pixel_size)]
        has_halfmaps = False

    grid_start = maps[0][1]
    g = maps[0][0]
    st.spacegroup_hm = "P1"
    st.cell = g.unit_cell

    if st[0].count_hydrogen_sites() > 0:
        monlib = utils.restraints.load_monomer_library(st[0].get_all_residue_names(),
                                                       monomer_dir=args.monlib)
    else:
        monlib = None

    if args.mask:
        mask = numpy.array(utils.fileio.read_ccp4_map(args.mask)[0])
    elif args.mask_radius:
        mask = gemmi.FloatGrid(*g.shape)
        mask.set_unit_cell(g.unit_cell)
        mask.spacegroup = gemmi.SpaceGroup(1)
        mask.mask_points_in_constant_radius(st[0], args.mask_radius, 1.)
        ccp4 = gemmi.Ccp4Map()
        ccp4.grid = mask
        ccp4.update_ccp4_header(2, True) # float, update stats
        ccp4.write_ccp4_map("mask_from_model.ccp4")
        mask = numpy.array(mask)
    else:
        mask = None
        if args.normalized_map:
            logger.write("Error: Provide --mask or --mask_radius if you want --normalized-map.")
            return

    hkldata, map_labs, stats_str = calc_fofc(st, args.resolution, maps, mask=mask, monlib=monlib, B=args.B,
                                             half1_only=args.half1_only, omit_proton=args.omit_proton,
                                             omit_h_electron=args.omit_h_electron)
    write_files(hkldata, map_labs, grid_start, stats_str,
                mask=mask, output_prefix=args.output_prefix,
                crop=args.crop, normalize_map=args.normalized_map, omit_h_electron=args.omit_h_electron)
    
# main()

if __name__ == "__main__":
    import sys
    args = parse_args(sys.argv[1:])
    main(args)



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import numpy
import os
import subprocess
import pipes
from servalcat.utils import logger
from servalcat import utils
from servalcat import spa

def add_arguments(parser):
    parser.description = 'FSC calculation'

    parser.add_argument('--model',
                        required=True,
                        help="")
    parser.add_argument('--maps',
                        required=True,
                        nargs="+",
                        help='Input map file(s)')
    parser.add_argument('--pixel_size', type=float,
                        help='Override pixel size (A)')
    parser.add_argument('-r', '--mask_radius',
                        type=float,
                        help='')
    parser.add_argument('-d', '--resolution',
                        type=float,
                        required=True,
                        help='')
    parser.add_argument('--fsc_out',
                        default="fsc.dat",
                        help='')
# add_arguments()
                        
def parse_args(arg_list):
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    return parser.parse_args(arg_list)
# parse_args()

def read_and_fft_maps(filenames, d_min, mask=None, pixel_size=None, check_consistency=True):
    hkldata = None

    last_cell, last_shape, last_sg = None, None, None
    
    for idx, mapin in enumerate(filenames):
        m = utils.fileio.read_ccp4_map(mapin, pixel_size=pixel_size)[0]
        if check_consistency:
            if last_cell is not None:
                assert m.unit_cell == last_cell
                assert m.shape == last_shape
                assert m.spacegroup == last_sg
            last_cell, last_shape, last_sg = m.unit_cell, m.shape, m.spacegroup

        if mask is not None:
            m = gemmi.FloatGrid(numpy.array(m) * mask, m.unit_cell, m.spacegroup)
            
        asu = gemmi.transform_map_to_f_phi(m).prepare_asu_data(dmin=d_min)
        label = "Fmap{}".format(idx+1)
        if hkldata is None:
            df = utils.hkl.df_from_asu_data(asu, label)
            hkldata = utils.hkl.HklData(m.unit_cell, m.spacegroup, False, df)
        else:
            hkldata.merge_asu_data(asu, label)
            
    return hkldata
# read_and_fft_maps()


def main(args):
    st = utils.fileio.read_structure(args.model)
    ref_grid = utils.fileio.read_ccp4_map(args.maps[0], pixel_size=args.pixel_size)[0]
    st.cell = ref_grid.unit_cell
    st.spacegroup_hm = "P1"

    if len(st.ncs) > 0:
        utils.model.expand_ncs(st)
    
    if args.mask_radius is not None:
        mask = gemmi.FloatGrid(*ref_grid.shape)
        mask.set_unit_cell(ref_grid.unit_cell)
        mask.spacegroup = ref_grid.spacegroup
        mask.mask_points_in_constant_radius(st[0], args.mask_radius, 1.)    
    else:
        mask = None
    
    fc = utils.model.calc_fc_fft(st, args.resolution, source="electron")
    hkldata = read_and_fft_maps(args.maps, args.resolution, mask, pixel_size=args.pixel_size)
    
    hkldata.merge_asu_data(fc, "FC")
    hkldata.setup_relion_binning()
    #print(hkldata.df)


    ofs = open(args.fsc_out, "w")
    ofs.write("# Mask_radius= {}\n".format(args.mask_radius))
    for i, f in enumerate(args.maps): ofs.write("# Map{} = {}\n".format(i+1, f))
    ofs.write("bin n d_max d_min {}\n".format(" ".join(["Fmap{}".format(i+1) for i in range(len(args.maps))])))
    fsc_list = [[] for i in range(len(args.maps))]
    n_list = [[] for i in range(len(args.maps))]
    for i_bin, bin_d_max, bin_d_min in hkldata.bin_and_limits():
        sel = i_bin == hkldata.df.bin
        fc = hkldata.df["FC"][sel]
        ofs.write("{:3d} {:7d} {:7.3f} {:7.3f}".format(i_bin, fc.size, bin_d_max, bin_d_min))
        for i_map in range(len(args.maps)):
            fo = hkldata.df["Fmap{}".format(i_map+1)][sel]
            fsc = numpy.real(numpy.corrcoef(fo, fc)[1,0])
            ofs.write(" {:.4f}".format(fsc))
            fsc_list[i_map].append(fsc)
            n_list[i_map].append(fc.size)
        ofs.write("\n")
    ofs.write("\n")
    ofs.write("#              FSCaverage =")
    for i in range(len(args.maps)):
        fsc = numpy.array(fsc_list[i])
        n = numpy.array(n_list[i])
        fsca = sum(fsc*n)/sum(n)
        ofs.write(" {:.4f}".format(fsca))
    ofs.write("\n")
# main()

if __name__ == "__main__":
    import sys
    args = parse_args(sys.argv[1:])
    main(args)



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
def parse_ncsc_keywords(kwd_str):
    # FIXME handle lines ending with -
    lines = kwd_str.splitlines()
    ret = []
    for l in lines:
        l = l.split()
        if len(l) < 3: continue
        if l[0].lower().startswith("ncsc"):
            if l[1].lower().startswith("matr"):
                vals = [float(x) for x in l[2:]]
                if len(vals) != 12:
                    print("Bad nsc matrix line: {}".format(" ".join(l)))
                    continue
                op = gemmi.NcsOp()
                op.tr.mat.fromlist([vals[3*x:3*x+3] for x in range(3)])
                op.tr.vec.fromlist(vals[9:])
                op.id = str(len(ret)+1)
                op.given = op.tr.is_identity()
                ret.append(op)
            elif l[1].lower().startswith("eule"):
                pass # TODO
            elif l[1].lower().startswith("pola"):
                pass # TODO
    return ret
# parse_ncsc_keywords()



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import subprocess
import pipes
import json
import copy
import re
import os
from servalcat.utils import logger

re_version = re.compile("#.* Refmac *version ([^ ]+) ")
re_error = re.compile('(warn|error *[:]|error *==|^error)', re.IGNORECASE)
re_outlier_start = re.compile("\*\*\*\*.*outliers")

def check_version(exe="refmac5"):
    p = subprocess.Popen([exe], shell=False, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                         universal_newlines=True)
    p.stdin.write("end\n")
    p.stdin.close()
    ver = ()
    for l in iter(p.stdout.readline, ""):
        r_ver = re_version.search(l)
        if r_ver:
            ver = tuple(map(int, r_ver.group(1).split(".")))
    p.wait()
    return ver
# check_version()

def external_restraints_json_to_keywords(json_in):
    ret = []
    exte_list = json.load(open(json_in))
    for e in exte_list:
        if "use" in e:
            ret.append("EXTERNAL USE {}".format(e["use"]))
        if "dmax" in e:
            ret.append("EXTERNAL DMAX {0}".format(e["dmax"]))
        if "weight_scale" in e:
            ret.append("EXTERNAL WEIGHT SCALE {0}".format(e["weight_scale"]))
        if "weight_gmwt" in e:
            ret.append("EXTERNAL WEIGHT GMWT {0}".format(e["weight_gmwt"]))
        if "file" in e:
            ret.append("@"+e["file"])

    return "\n".join(ret) + "\n"
# external_restraints_json_to_keywords()

class Refmac:
    def __init__(self, **kwargs):
        self.prefix = "refmac"
        self.hklin = self.xyzin = ""
        self.lab_f = None
        self.lab_sigf = None
        self.lab_phi = None
        self.hydrogen = "all"
        self.hout = "no"
        self.ncycle = 10
        self.resolution = None
        self.weight_fixed = None
        self.weight_auto_scale = None
        self.bfactor = None
        self.jellybody = None
        self.jellybody_sigma, self.jellybody_dmax = 0.01, 4.2
        self.ncsr = None
        self.keyword_files = []
        self.keywords = []
        self.external_restraints_json = None
        self.exe = "refmac5"
        self.monlib_path = None
        self.show_log = False # summary only if false
        self.global_mode = kwargs.get("global_mode")
        
        if self.global_mode == "spa":
            self.source = "em mb"
            self.solvent = "no"
        else:
            self.source = "xray"
            self.solvent = "yes"
            
        for k in kwargs:
            if k == "args":
                self.init_from_args(kwargs["args"])
            else:
                setattr(self, k, kwargs[k])
    # __init__()

    def init_from_args(self, args):
        self.hklin = args.mtz
        self.xyzin = args.model
        self.libin = args.ligand
        self.ncycle = args.ncycle
        self.lab_f = args.lab_f
        self.lab_phi = args.lab_phi
        self.lab_sigf = args.lab_sigf
        self.hydrogen = args.hydrogen
        self.hout = args.hout
        self.ncsr = args.ncsr
        self.bfactor = args.bfactor
        self.jellybody = args.jellybody
        self.jellybody_sigma, self.jellybody_dmax = args.jellybody_params
        self.resolution = args.resolution
        self.weight_auto_scale = args.weight_auto_scale
        self.weight_fixed = args.weight
        self.keyword_files = args.keyword_file
        self.keywords = args.keywords
        self.external_restraints_json = args.external_restraints_json
        self.exe = args.exe
        self.show_log = args.show_refmac_log
        self.monlib_path = args.monlib
    # init_from_args()

    def copy(self, **kwargs):
        ret = copy.copy(self)
        for k in kwargs:
            setattr(ret, k, kwargs[k])
            
        return ret
    # copy()

    def set_libin(self, ligands):
        if not ligands: return
        if len(ligands) > 1:
            mcif = "merged_ligands.cif" # XXX directory!
            logger.write("Merging ligand cif files: {}".format(ligands))
            utils.fileio.merge_ligand_cif(ligands, mcif)
            self.libin = mcif
        else:
            self.libin = ligands[0]
    # set_libin()
            
    
    def make_keywords(self):
        ret = ""
        labin = []
        if self.lab_f: labin.append("FP={}".format(self.lab_f))
        if self.lab_sigf: labin.append("SIGFP={}".format(self.lab_sigf))
        if self.lab_phi: labin.append("PHIB={}".format(self.lab_phi))
        ret += "labin {}\n".format(" ".join(labin))


        ret += "make hydr {}\n".format(self.hydrogen)
        ret += "make hout {}\n".format("yes" if self.hout else "no")
        ret += "solvent {}\n".format(self.solvent)
        ret += "source {}\n".format(self.source)
        ret += "ncycle {}\n".format(self.ncycle)
        if self.resolution is not None:
            ret += "reso {}\n".format(self.resolution)
        if self.weight_fixed is not None:
            ret += "weight matrix {}\n".format(self.weight_fixed)
        elif self.weight_auto_scale is not None:
            ret += "weight auto {:.2f}\n".format(self.weight_auto_scale)
        else:
            ret += "weight auto\n"

        if self.bfactor is not None:
            ret += "bfactor set {}\n".format(self.bfactor)
        if self.jellybody:
            ret += "ridge dist sigma {:.3f}\n".format(self.jellybody_sigma)
            ret += "ridge dist dmax {:.2f}\n".format(self.jellybody_dmax)
        if self.ncsr:
            ret += "ncsr {}\n".format(self.ncsr)

        if self.external_restraints_json:
            ret += external_restraints_json_to_keywords(self.external_restraints_json)
            
        if self.keyword_files:
            for f in self.keyword_files:
                ret += "@{}\n".format(f)

        if self.keywords:
            ret += "\n".join(self.keywords).strip() + "\n"

        return ret
    # make_keywords()

    def xyzout(self): return self.prefix + ".pdb"
    def hklout(self): return self.prefix + ".mtz"

    def make_cmd(self):
        cmd = [self.exe]
        cmd.extend(["hklin", pipes.quote(self.hklin)])
        cmd.extend(["hklout", pipes.quote(self.hklout())])
        cmd.extend(["xyzin", pipes.quote(self.xyzin)])
        cmd.extend(["xyzout", pipes.quote(self.xyzout())])
        if self.libin:
            cmd.extend(["libin", self.libin])

        return cmd
    # make_cmd()

    def run_refmac(self):
        cmd = self.make_cmd()
        stdin = self.make_keywords()
        open(self.prefix+".inp", "w").write(stdin)

        logger.write("Running REFMAC5..")
        logger.write("{} <<__eof__ > {}".format(" ".join(cmd), self.prefix+".log"))
        logger.write(stdin, end="")
        logger.write("__eof__")

        env = os.environ
        if self.monlib_path: env["CLIBD_MON"] = os.path.join(self.monlib_path, "") # should end with /
        
        p = subprocess.Popen(cmd, shell=False, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                             universal_newlines=True, env=env)
        p.stdin.write(stdin)
        p.stdin.close()

        log = open(self.prefix+".log", "w")
        cycle = 0
        re_lastcycle = re.compile("Cycle *{}. Rfactor analysis".format(self.ncycle+1))
        rmsbond = ""
        rmsangle = ""
        log_delay = []
        summary_write = (lambda x: log_delay.append(x)) if self.show_log else logger.write
        outlier_flag = False
        
        for l in iter(p.stdout.readline, ""):
            log.write(l)

            if self.show_log:
                print(l, end="")
            
            r_ver = re_version.search(l)
            if r_ver:
                summary_write("Starting Refmac {}".format(r_ver.group(1)))

            # print error/warning
            r_err = re_error.search(l)
            if r_err:
                if self.global_mode == "spa":
                    if "Figure of merit of phases has not been assigned" in l:
                        continue
                    elif "They will be assumed to be equal to 1.0" in l:
                        continue
                summary_write(l.rstrip())

            # print outliers
            r_outl = re_outlier_start.search(l)
            if r_outl:
                outlier_flag = True
                summary_write(l.rstrip())
            elif outlier_flag:
                if l.strip() == "" or "monitored" in l or "dev=" in l or "sigma=" in l.lower() or "sigma.=" in l:
                    summary_write(l.rstrip())
                else:
                    outlier_flag = False
                    
            if self.global_mode == "spa":
                if "CGMAT cycle number =" in l:
                    cycle = int(l[l.index("=")+1:])
                elif re_lastcycle.search(l):
                    cycle = self.ncycle + 1
                elif "Average Fourier shell correlation    =" in l and cycle > 0:
                    fsc = l[l.index("=")+1:].strip()
                    if cycle == 1:
                        note = "(initial)"
                    elif cycle > self.ncycle:
                        note = "(final)"
                    else:
                        note = ""
                        
                    summary_write(" cycle= {:3d} FSCaverage= {} {}".format(cycle-1, fsc, note))
                    cycle = 0
                elif "Rms BondLength" in l:
                    rmsbond = l
                elif "Rms BondAngle" in l:
                    rmsangle = l

        ret = p.wait()
        if log_delay:
            logger.write("== Summary of Refmac ==")
            logger.write("\n".join(log_delay))
                         
        if rmsbond:
            logger.write("                      Initial    Final")
            logger.write(rmsbond.rstrip())
            logger.write(rmsangle.rstrip())
        logger.write("REFMAC5 finished with exit code= {}".format(ret))

        if not os.path.isfile(self.xyzout()) or not os.path.isfile(self.hklout()):
            raise RuntimeError("REFMAC5 did not produce output files.")
        
    # run_refmac()
# class Refmac



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import numpy
import argparse
from servalcat.utils import logger
from servalcat import utils
from servalcat.spa import shift_maps

def add_sfcalc_args(parser):
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--halfmaps", nargs=2, help="Input half map files")
    group.add_argument("--map", help="Use only if you really do not have half maps.")
    parser.add_argument('--mapref',
                        help='Reference map file')
    parser.add_argument('--mask',
                        help='Mask file')
    parser.add_argument('--model',
                        help='Input atomic model file')
    parser.add_argument('--mask_radius',
                        type=float, default=3,
                        help='')
    parser.add_argument('--padding',
                        type=float, 
                        help='Default: 2*mask_radius')
    parser.add_argument('--no_mask',
                        action='store_true')
    parser.add_argument('--pixel_size', type=float,
                        help='Override pixel size (A)')
    parser.add_argument('--resolution',
                        type=float,
                        help='')
    parser.add_argument('--no_shift',
                        action='store_true',
                        help='')
    parser.add_argument('--blur',
                        nargs="+", # XXX probably no need to be multiple
                        type=float,
                        help='Sharpening or blurring B')
    parser.add_argument('--pg',
                        help="Point group symbol for strict symmetry. The coordinate system is consitent with RELION.")
    parser.add_argument('--twist', type=float, help="Helical twist (degree)")
    parser.add_argument('--rise', type=float, help="Helical rise (Angstrom)")
    parser.add_argument('--ignore_symmetry',
                        help='Ignore symmetry information in the model file')
    parser.add_argument('--remove_multiple_models', action='store_true', 
                        help='Keep 1st model only')
    parser.add_argument("--b_before_mask", type=float)
    parser.add_argument('--no_sharpen_before_mask', action='store_true',
                        help='By default half maps are sharpened before masking by std of signal and unsharpened after masking. This option disables it.')
    parser.add_argument('--no_fix_microheterogeneity', action='store_true', 
                        help='By default it will fix microheterogeneity for Refmac')

# add_sfcalc_args()

def add_arguments(parser):
    parser.description = 'Structure factor calculation for Refmac'
    add_sfcalc_args(parser)
    parser.add_argument('--output_masked_prefix',
                        default="masked_fs",
                        help='output MTZ file name for masked F')
    parser.add_argument('--output_mtz_prefix',
                        default="starting_map",
                        help='output MTZ file name for original F')
    parser.add_argument('--output_model_prefix',
                        default="shifted_local",
                        help='output model file name')
# add_arguments()

def parse_args(arg_list):
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    return parser.parse_args(arg_list)
# parse_args()

def write_map_mtz(hkldata, mtz_out, map_labs, sig_lab=None, blurs=None):
    if not blurs: blurs = []
    if 0 not in blurs: blurs = [0.] + blurs
    
    nblur = len(blurs)
    ncol = 3+len(map_labs)*(1+nblur)
    if sig_lab: ncol += nblur
    data = numpy.empty((len(hkldata.df.index), ncol))
    data[:,:3] = hkldata.df[["H","K","L"]]
    s2 = 1./hkldata.d_spacings()**2
    for i, lab in enumerate(map_labs):
        for j, b in enumerate(blurs):
            f = numpy.abs(hkldata.df[lab])
            if b != 0: f *= numpy.exp(-b*s2/4.)
            data[:,3+i*(1+nblur)+j] = f
        data[:,3+i*(1+nblur)+nblur] = numpy.angle(hkldata.df[lab], deg=True)
        
    if sig_lab:
        for j, b in enumerate(blurs):
            sigf = hkldata.df[sig_lab]
            if b != 0: sigf *= numpy.exp(-b*s2/4.)
            data[:,3+len(map_labs)*(1+nblur)+j] = sigf

    mtz = gemmi.Mtz()
    mtz.spacegroup = hkldata.sg
    mtz.cell = hkldata.cell
    mtz.add_dataset('HKL_base')
    for label in ['H', 'K', 'L']: mtz.add_column(label, 'H')

    for lab in map_labs:
        lab_root = lab[1:] if lab[0]=="F" else lab
        for b in blurs:
            labf = lab
            if b != 0: labf = "{}Blur_{:.2f}".format(labf, b).replace("Fout0", "Fout")
            mtz.add_column(labf, "F")
        mtz.add_column("P"+lab_root, "P")
    if sig_lab:
        for b in blurs:
            labsigf = sig_lab
            if b != 0: labsigf = "{}Blur_{:.2f}".format(labsigf, b).replace("Fout0", "Fout")
            mtz.add_column(labsigf, "Q")
        
    mtz.set_data(data)
    mtz.write_to_file(mtz_out)
# write_map_mtz()

def scale_maps(maps_in, map_ref, d_min):
    fs = []
    for m in [[map_ref,None]]+maps_in:
        asu = gemmi.transform_map_to_f_phi(m[0]).prepare_asu_data(dmin=d_min)
        fs.append(asu)

    binner = utils.hkl.Binner(fs[0], style="relion")
    #fs, binner = fft_and_binning([map_ref]+maps_in, d_min)
    f_ref, f_inps = fs[0], fs[1:]
    d_array = f_ref.make_d_array()
    f_inps_scaled = [f_obs.copy() for f_obs in f_inps]
    logger.write(" Saving sf stats as f_stats.dat")
    ofs = open("f_stats.dat", "w")
    lab_fos = " ".join(["F_obs.{n} F_scaled.{n}".format(n=i+1) for i in range(len(f_inps))])
    ofs.write("bin count d_min F_ref {}\n".format(lab_fos))
    for i_bin, c_bin in zip(binner.bins, binner.bin_counts):
        sel = binner.bin_array == i_bin
        fr = f_ref.value_array[sel]
        s_fr = numpy.std(fr)
        d = min(d_array[sel])
        avg_strs = []
        for f_obs, f_obs_scaled in zip(f_inps, f_inps_scaled):
            fo = f_obs.value_array[sel]
            s_fo = numpy.std(fo)
            fo_scaled = fo*s_fr/s_fo
            f_obs_scaled.value_array[sel] = fo_scaled
            avg_strs.append("{:e} {:e}".format(numpy.average(numpy.abs(fo)), numpy.average(numpy.abs(fo_scaled))))

        ofs.write("{:3d} {:6d} {:7.3f} {:e} {}\n".format(i_bin, c_bin, d,
                                                         numpy.average(numpy.abs(fr)), " ".join(avg_strs)))
                                                                

    maps_scaled = [gemmi.transform_f_phi_grid_to_map(f.get_f_phi_on_grid(size=map_ref.shape))
                   for f in f_inps_scaled]
    return [[x]+y[1:] for x,y in zip(maps_scaled, maps_in)]
# scale_maps()

def main(args, monlib=None):
    ret = {} # instructions for refinement
    
    if (args.twist, args.rise).count(None) == 1:
        raise RuntimeError("ERROR: give both helical paramters --twist and --rise")

    is_helical = args.twist is not None

    if args.no_mask:
        args.mask_radius = None
        if not args.no_shift:
            logger.write("WARNING: setting --no_shift because --no_mask is given")
            args.no_shift = True
        if args.mask:
            logger.write("WARNING: Your --mask is ignored because --no_mask is given")
            args.mask = None

    if args.resolution is None and args.model and utils.fileio.splitext(args.model)[1].endswith("cif"):
        doc = gemmi.cif.read(args.model)
        block = doc.sole_block()
        reso_str = block.find_value("_em_3d_reconstruction.resolution")
        logger.write("WARNING: --resolution not given. Using _em_3d_reconstruction.resolution = {}".format(reso_str))
        args.resolution = float(reso_str)

    if args.resolution is None:
        raise RuntimeError("ERROR: --resolution is needed.")
        
    resolution = args.resolution - 1e-6

    if args.halfmaps:
        maps = [utils.fileio.read_ccp4_map(f, pixel_size=args.pixel_size) for f in args.halfmaps]
        assert maps[0][0].shape == maps[1][0].shape
        assert maps[0][0].unit_cell == maps[1][0].unit_cell
        assert maps[0][1] == maps[1][1]
    else:
        maps = [utils.fileio.read_ccp4_map(args.map, pixel_size=args.pixel_size)]

    grid_start = maps[0][1]
    unit_cell = maps[0][0].unit_cell
    spacegroup = gemmi.SpaceGroup(1)
    start_xyz = numpy.array(maps[0][0].get_position(*grid_start).tolist())
    A = numpy.array(unit_cell.orthogonalization_matrix.tolist())
    center = numpy.sum(A, axis=1) / 2 + start_xyz

    if args.mapref:
        logger.write("Reference map: {}".format(args.mapref))
        map_ref, ref_start = utils.fileio.read_ccp4_map(args.mapref, pixel_size=args.pixel_size)
        assert unit_cell == map_ref.unit_cell
        assert maps[0][0].shape == map_ref.shape
        assert maps[0][1] == ref_start

        # Overwrite input_maps
        logger.write("Scaling maps..")
        maps = scale_maps(maps, map_ref, resolution)

    # Create mask
    mask = None
    if args.mask:
        logger.write("Input mask file: {}".format(args.mask))
        mask = numpy.array(utils.fileio.read_ccp4_map(args.mask)[0])
    
    st_new = None
    if args.model: # and 
        logger.write("Input model: {}".format(args.model))
        st = utils.fileio.read_structure(args.model)
        st.cell = unit_cell
        st.spacegroup_hm = "P 1"

        if not args.no_fix_microheterogeneity:
            # TODO need to check external restraints

            if monlib is None:
                # FIXME should use user provided libraries
                monlib = utils.restraints.load_monomer_library(st[0].get_all_residue_names())
            mhtr_mods = utils.model.microheterogeneity_for_refmac(st, monlib)
            ret["inscode_mods"] = mhtr_mods
            
        model_format = utils.fileio.check_model_format(args.model)
        chain_id_len_max = max([len(x) for x in utils.model.all_chain_ids(st)])
        if chain_id_len_max > 1 and model_format == ".pdb":
            logger.write("Long chain ID (length: {}) detected. Will use mmcif format".format(chain_id_len_max))
            model_format = ".mmcif"

        ret["model_format"] = model_format

        if args.remove_multiple_models and len(st) > 1:
            logger.write(" Removing models 2-{}".format(len(st)))
            for i in reversed(range(1, len(st))):
                del st[i]

        if args.ignore_symmetry and len(st.ncs) > 0:
            logger.write("Removing symmetry information from model.")
            st.ncs.clear()

        if is_helical:
            ops = utils.symmetry.generate_helical_operators(st, start_xyz, center,
                                                            args.pg, args.twist, args.rise)
            logger.write("{} helical operators found".format(len(ops)))
            st.ncs.clear()
            st.ncs.extend([x for x in ops if not x.tr.is_identity()])
            #ret["helical"] = ops
            utils.model.filter_helical_contacting(st)
        elif args.pg:
            logger.write("Point group symmetry: {}".format(args.pg))
            if len(st.ncs) > 0:
                logger.write(" WARNING: NCS information in model file will be ignored")

            _, _, ops = utils.symmetry.operators_from_symbol(args.pg)
            if ops:
                logger.write(" {} operators found".format(len(ops)))
                ops = utils.symmetry.make_NcsOps_from_matrices(ops, cell=unit_cell, center=center)
                st.ncs.clear()
                st.ncs.extend([x for x in ops if not x.tr.is_identity()])

        elif len(st.ncs) > 0:
            logger.write("Strict NCS detected from model.")

        if len(st.ncs) > 0:
            logger.write(" Writing NCS file")
            utils.symmetry.write_NcsOps_for_refmac(st.ncs, "ncsc_global.txt")
            ret["ncsc_file"] = "ncsc_global.txt"
        
        st_new = st.clone()
        if len(st.ncs) > 0 and not args.no_shift:
            utils.model.expand_ncs(st)
            logger.write(" Saving expanded model: input_model_expanded.*")
            utils.fileio.write_model(st, "input_model_expanded", pdb=True, cif=True)
    
        if not mask and args.mask_radius:
            logger.write("Creating mask..")
            mask = gemmi.FloatGrid(*maps[0][0].shape)
            mask.set_unit_cell(unit_cell)
            mask.spacegroup = spacegroup
            mask.mask_points_in_constant_radius(st[0], args.mask_radius, 1.)
            ccp4 = gemmi.Ccp4Map()
            ccp4.grid = mask
            ccp4.update_ccp4_header(2, True) # float, update stats
            ccp4.write_ccp4_map("mask_from_model.ccp4")
            logger.write("Mask file written: mask_from_model.ccp4")
    else:
        model_format = None
        
    if args.no_shift:
        logger.write(" Saving input model with unit cell information")
        utils.fileio.write_model(st, "starting_model", pdb=True, cif=True)
        ret["model_file"] = "starting_model" + model_format

    if mask:
        # Mask maps
        if args.no_sharpen_before_mask or len(maps) < 2:
            logger.write("Applying mask..")
            maps = [[gemmi.FloatGrid(numpy.array(ma[0])*mask, unit_cell, spacegroup)]+ma[1:]
                    for ma in maps]
        else:
            logger.write("Sharpen-mask-unsharpen..")
            maps = utils.maps.sharpen_mask_unsharpen(maps, mask, resolution, b=args.b_before_mask)

        if not args.no_shift:
            logger.write(" Shifting maps and/or model..")
            if args.padding is None: args.padding = args.mask_radius * 2
            new_cell, new_shape, starts, shifts = shift_maps.determine_shape_and_shift(mask=mask,
                                                                                       grid_start=grid_start,
                                                                                       padding=args.padding,
                                                                                       mask_cutoff=0.1,
                                                                                       noncentered=True,
                                                                                       noncubic=True,
                                                                                       json_out="shifts.json")
            ret["shifts"] = shifts
            vol_mask = numpy.count_nonzero(numpy.array(mask)>0.5)
            vol_map = new_shape[0] * new_shape[1] * new_shape[2]
            ret["vol_ratio"] = vol_mask/vol_map
            logger.write(" Vol_mask/Vol_map= {:.2e}".format(ret["vol_ratio"]))
            
            if st_new:
                for cra in st_new[0].all():
                    cra.atom.pos += shifts
                
                st_new.cell = new_cell
                st_new.spacegroup_hm = "P 1"
                if len(st_new.ncs) > 0:
                    new_ops = utils.symmetry.apply_shift_for_ncsops(st_new.ncs, shifts)
                    st_new.ncs.clear()
                    st_new.ncs.extend(new_ops)
                    logger.write(" Writing NCS file for shifted model")
                    utils.symmetry.write_NcsOps_for_refmac(st_new.ncs, "ncsc_local.txt")
                    ret["ncsc_file"] = "ncsc_local.txt"
                    logger.write(" Writing symmetry expanded model for shifted model")
                    utils.symmetry.write_symmetry_expanded_model(st_new, "shifted_local_expanded",
                                                                 pdb=True, cif=True)

                logger.write(" Saving shifted model..")
                utils.fileio.write_model(st_new, "shifted_local", pdb=True, cif=True)
                ret["model_file"] = "shifted_local" + model_format

            logger.write(" Trimming maps..")
            for i in range(len(maps)): # Update maps
                suba = maps[i][0].get_subarray(*(list(starts)+list(new_shape)))
                new_grid = gemmi.FloatGrid(suba, new_cell, spacegroup)
                maps[i][0] = new_grid

        
    hkldata = utils.maps.mask_and_fft_maps(maps, resolution, None)
    hkldata.setup_relion_binning()
    if len(maps) == 2:
        logger.write(" Calculating noise variances..")
        map_labs = ["Fmap1", "Fmap2", "Fout0"]
        sig_lab = "SIGFout0"
        ret["lab_sigf"] = sig_lab
        ret["lab_f_half1"] = "Fmap1"
        # TODO Add SIGF in case of half maps, when refmac is ready
        ret["lab_phi_half1"] = "Pmap1"
        ret["lab_f_half2"] = "Fmap2"
        ret["lab_phi_half2"] = "Pmap2"
        utils.maps.calc_noise_var_from_halfmaps(hkldata)
        hkldata.df[sig_lab] = 0.
        for i_bin, bin_d_max, bin_d_min in hkldata.bin_and_limits():
            sel = i_bin == hkldata.df.bin
            hkldata.df.loc[sel, sig_lab] = numpy.sqrt(hkldata.binned_df["var_noise"][i_bin])

        d_eff_full = hkldata.d_eff("FSCfull")
        logger.write("Effective resolution from FSCfull= {:.2f}".format(d_eff_full))
        ret["d_eff"] = d_eff_full
    else:
        map_labs = ["Fout0"]
        sig_lab = None

    if args.no_shift:
        logger.write("Saving original maps as mtz files..")
        mtzout = args.output_mtz_prefix+".mtz"
    else:
        logger.write(" Saving masked maps as mtz files..")
        mtzout = args.output_masked_prefix+"_obs.mtz"

    hkldata.df.rename(columns=dict(F_map1="Fmap1", F_map2="Fmap2", FP="Fout0"), inplace=True)
    write_map_mtz(hkldata, mtzout,
                  map_labs=map_labs, sig_lab=sig_lab, blurs=args.blur)
    ret["mtz_file"] = mtzout
    ret["lab_f"] = "Fout" + ("Blur_{:.2f}".format(args.blur[0]) if args.blur else "0")
    ret["lab_phi"] = "Pout0"
    return ret
# main()

if __name__ == "__main__":
    import sys
    args = parse_args(sys.argv[1:])
    main(args)



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
from servalcat.utils import logger
from servalcat.utils import restraints
import gemmi
import numpy
import scipy.sparse
import os
import itertools
import string

def shake_structure(st, sigma):
    print("Randomizing structure with rmsd of {}".format(sigma))
    st2 = st.clone()
    sigma /= numpy.sqrt(3)
    for model in st2:
        for cra in model.all():
            r = numpy.random.normal(0, sigma, 3)
            cra.atom.pos += gemmi.Position(*r)

    return st2
# shake_structure()

_normalized_it92_elems = set()

def normalize_it92(st=None, all_elements=False, quiet=False):
    elements = set()
        
    if all_elements:
        elements.add("H")
        elements.add("D")
        for i in range(2, 119):
            elements.add(gemmi.Element(i).name)
    elif st is not None:
        for cra in st[0].all():
            elements.add(cra.atom.element.name)

    for el in elements:
        elem = gemmi.Element(el)
        if elem.name in _normalized_it92_elems: continue
        z = elem.atomic_number
        coef = elem.it92
        if coef is None:
            logger.write("IT92 table not found for {}".format(elem.name))
            continue
        norm = z/(sum(coef.a)+coef.c)
        if not quiet: logger.write("Normalizing atomic scattering factor of {} by {}".format(el, norm))
        new_coef = [x*norm for x in coef.a] + coef.b + [coef.c*norm]
        coef.set_coefs(new_coef)
        _normalized_it92_elems.add(elem.name)
# normalize_it92()

def determine_blur_for_dencalc(st, grid):
    b_min = min((cra.atom.b_iso for cra in st[0].all()))
    b_need = grid**2*8*numpy.pi**2/1.1 # Refmac's way
    b_add = b_need - b_min
    return b_add
# determine_blur_for_dencalc()

def calc_fc_fft(st, d_min, source, mott_bethe=True, monlib=None, blur=None, cutoff=1e-5, rate=1.5,
                omit_proton=False, omit_h_electron=False):
    assert source in ("xray", "electron")
    normalize_it92(st)
    if source != "electron": assert not mott_bethe
    if omit_proton or omit_h_electron:
        assert mott_bethe
        if st[0].count_hydrogen_sites() == 0:
            logger.write("WARNING: omit_proton/h_electron requested, but no hydrogen exists!")
            omit_proton = omit_h_electron = False
        elif omit_proton and omit_h_electron:
            logger.write("omit_proton and omit_h_electron requested. removing hydrogens")
            st = st.clone()
            st.remove_hydrogens()
            omit_proton = omit_h_electron = False
    
    if blur is None: blur = determine_blur_for_dencalc(st, d_min/2/rate)
    blur = max(0, blur) # negative blur may cause non-positive definite in case of anisotropic Bs
    logger.write("Setting blur= {:.2f} in density calculation".format(blur))
        
    if mott_bethe and not omit_proton and monlib is not None and st[0].count_hydrogen_sites() > 0:
        st = st.clone()
        topo = gemmi.prepare_topology(st, monlib)
        resnames = st[0].get_all_residue_names()
        restraints.check_monlib_support_nucleus_distances(monlib, resnames)
        # Shift electron positions
        topo.adjust_hydrogen_distances(gemmi.Restraints.DistanceOf.ElectronCloud)
    else:
        topo = None
        
    if source == "xray" or mott_bethe:
        dc = gemmi.DensityCalculatorX()
    else:
        dc = gemmi.DensityCalculatorE()

    dc.d_min = d_min
    dc.blur = blur
    dc.cutoff = cutoff
    dc.rate = rate
    dc.set_grid_cell_and_spacegroup(st)

    if mott_bethe:
        if omit_proton:
            method_str = "proton-omit Fc"
        elif omit_h_electron:
            if topo is None:
                method_str = "hydrogen electron-omit Fc"
            else:
                method_str = "hydrogen electron-omit, proton-shifted Fc"
        elif topo is not None:
            method_str = "proton-shifted Fc"
        else:
            method_str = "Fc"

        logger.write("Calculating {} using Mott-Bethe formula".format(method_str))
            
        dc.initialize_grid()
        dc.addends.subtract_z(except_hydrogen=True)

        if omit_h_electron:
            st2 = st.clone()
            st2.remove_hydrogens()
            dc.add_model_density_to_grid(st2[0])
        else:
            dc.add_model_density_to_grid(st[0])

        # Subtract hydrogen Z
        if not omit_proton and st[0].count_hydrogen_sites() > 0:
            if topo is not None:
                # Shift proton positions
                topo.adjust_hydrogen_distances(gemmi.Restraints.DistanceOf.Nucleus,
                                               default_scale=restraints.default_proton_scale)
            for cra in st[0].all():
                if cra.atom.is_hydrogen():
                    dc.add_c_contribution_to_grid(cra.atom, -1)

        dc.grid.symmetrize_sum()
    else:
        logger.write("Calculating Fc")
        dc.put_model_density_on_grid(st[0])

    grid = gemmi.transform_map_to_f_phi(dc.grid)
    asu_data = grid.prepare_asu_data(dmin=d_min, mott_bethe=mott_bethe, unblur=dc.blur)
        
    return asu_data

# calc_fc_fft()

def calc_fc_direct(st, d_min, source, mott_bethe, monlib=None):
    assert source in ("xray", "electron")
    if source != "electron": assert not mott_bethe
    normalize_it92(st)

    unit_cell = st.cell
    spacegroup = gemmi.SpaceGroup(st.spacegroup_hm)
    miller_array = gemmi.make_miller_array(unit_cell, spacegroup, d_min)
    topo = None

    if source == "xray" or mott_bethe:
        calc = gemmi.StructureFactorCalculatorX(st.cell)
    else:
        calc = gemmi.StructureFactorCalculatorE(st.cell)
        
    
    if source == "electron" and mott_bethe:
        if monlib is not None and st[0].count_hydrogen_sites() > 0:
            st = st.clone()
            topo = gemmi.prepare_topology(st, monlib)
            resnames = st[0].get_all_residue_names()
            restraints.check_monlib_support_nucleus_distances(monlib, resnames)

        calc.addends.clear()
        calc.addends.subtract_z(except_hydrogen=True)

    vals = []
    for hkl in miller_array:
        sf = calc.calculate_sf_from_model(st[0], hkl)
        if mott_bethe: sf *= calc.mott_bethe_factor()
        vals.append(sf)

    if topo is not None:
        topo.adjust_hydrogen_distances(gemmi.Restraints.DistanceOf.Nucleus,
                                       default_scale=restraints.default_proton_scale)

        for i, hkl in enumerate(miller_array):
            sf = calc.calculate_mb_z(st[0], hkl, only_h=True)
            if mott_bethe: sf *= calc.mott_bethe_factor()
            vals[i] += sf
    
    asu = gemmi.ComplexAsuData(unit_cell, spacegroup,
                               miller_array, vals)
    return asu
# calc_fc_direct()

def get_em_expected_hydrogen(st, d_min, monlib=None, blur=None, cutoff=1e-5, rate=1.5):
    # Very crude implementation to find peak from calculated map
    # TODO Need to implement peak finding function that involves interpolation
    assert st[0].count_hydrogen_sites() > 0
    if blur is None: blur = determine_blur_for_dencalc(st, d_min/2/rate)
    blur = max(0, blur)
    logger.write("Setting blur= {:.2f} in density calculation".format(blur))

    st = st.clone()
    topo = gemmi.prepare_topology(st, monlib)
    resnames = st[0].get_all_residue_names()
    restraints.check_monlib_support_nucleus_distances(monlib, resnames)

    topo.adjust_hydrogen_distances(gemmi.Restraints.DistanceOf.ElectronCloud)
    st_e = st.clone()
    topo.adjust_hydrogen_distances(gemmi.Restraints.DistanceOf.Nucleus)
    st_n = st.clone()
    
    dc = gemmi.DensityCalculatorX()
    dc.d_min = d_min
    dc.blur = blur
    dc.cutoff = cutoff
    dc.rate = rate

    # Decide box_size
    max_r = max([dc.estimate_radius(cra.atom) for cra in st[0].all()])
    logger.write("max_r= {:.2f}".format(max_r))
    box_size = max_r*2 + 1 # padding
    logger.write("box_size= {:.2f}".format(box_size))
    mode_all = False #True
    if mode_all:
        dc.set_grid_cell_and_spacegroup(st)
    else:
        dc.grid.unit_cell = gemmi.UnitCell(box_size, box_size, box_size, 90, 90, 90)
        dc.grid.spacegroup = gemmi.SpaceGroup("P1")
        cbox = gemmi.Position(box_size/2, box_size/2, box_size/2)
    
    if mode_all: dc.initialize_grid()

    for ichain in range(len(st[0])):
        chain = st[0][ichain]
        for ires in range(len(chain)):
            residue = chain[ires]
            for iatom in range(len(residue)):
                atom = residue[iatom]
                if not atom.is_hydrogen(): continue
                h_n = st_n[0][ichain][ires][iatom]
                h_e = st_e[0][ichain][ires][iatom]
                if not mode_all:
                    dc.initialize_grid()
                    h_n.occ = 1.
                    h_e.occ = 1.
                    n_pos = gemmi.Position(h_n.pos)
                    h_n.pos = cbox
                    h_e.pos = cbox + h_e.pos - n_pos
                dc.add_atom_density_to_grid(h_e)
                dc.add_c_contribution_to_grid(h_n, -1)
                if not mode_all:
                    grid = gemmi.transform_map_to_f_phi(dc.grid)
                    asu_data = grid.prepare_asu_data(dmin=d_min, mott_bethe=True, unblur=dc.blur)
                    denmap = asu_data.transform_f_phi_to_map(exact_size=(int(box_size*10), int(box_size*10), int(box_size*10)))
                    m = numpy.unravel_index(numpy.argmax(denmap), denmap.shape)
                    peakpos = denmap.get_position(m[0], m[1], m[2]) - cbox + n_pos
                    atom.pos = peakpos

    if mode_all:
        grid = gemmi.transform_map_to_f_phi(dc.grid)
        asu_data = grid.prepare_asu_data(dmin=d_min, mott_bethe=True, unblur=dc.blur)
        denmap = asu_data.transform_f_phi_to_map(sample_rate=3)
        ccp4 = gemmi.Ccp4Map()
        ccp4.grid = denmap
        ccp4.update_ccp4_header(2, True) # float, update stats
        ccp4.write_ccp4_map("debug.ccp4")

    return st

# get_em_expected_hydrogen()

def translate_into_box(st):
    # apply unit cell translations to put model into a box (unit cell)
    omat = numpy.array(st.cell.orthogonalization_matrix)
    fmat = numpy.array(st.cell.fractionalization_matrix).transpose()
    for m in st:
        com = numpy.array(m.calculate_center_of_mass().tolist())
        shift = sum([omat[:,i]*numpy.floor(1-numpy.dot(com, fmat[:,i])) for i in range(3)])
        tr = gemmi.Transform(gemmi.Mat33(), gemmi.Vec3(*shift))
        m.transform(tr)
# translate_into_box()

def cra_to_indices(cra, model):
    ret = [None, None, None]
    for ic in range(len(model)):
        chain = model[ic]
        if cra.chain != chain: continue
        ret[0] = ic
        for ir in range(len(chain)):
            res = chain[ir]
            if cra.residue != res: continue
            ret[1] = ir
            for ia in range(len(res)):
                if cra.atom == res[ia]:
                    ret[2] = ia

    return tuple(ret)
# cra_to_indices()

def expand_ncs(st, special_pos_threshold=0.01):
    if len(st.ncs) == 0: return
    logger.write("Expanding symmetry..")
    st.expand_ncs(gemmi.HowToNameCopiedChain.Dup)

    # Take care of special positions
    if special_pos_threshold >= 0:
        cra2key = lambda x: (x.chain.name, x.residue.seqid.num, x.residue.seqid.icode,
                             x.atom.name, x.atom.element.name, x.atom.altloc.replace("\0"," "))
        ns = gemmi.NeighborSearch(st[0], st.cell, 3).populate()
        cs = gemmi.ContactSearch(special_pos_threshold)
        #cs.ignore = gemmi.ContactSearch.Ignore.SameAsu
        #cs.special_pos_cutoff_sq = special_pos_threshold
        results = cs.find_contacts(ns)

        # find overlaps between different segments
        pairs = {}
        cra_dict = {}
        for r in results:
            if r.partner1.residue.segment == r.partner2.residue.segment: continue
            key1, key2 = cra2key(r.partner1), cra2key(r.partner2)
            if key1 == key2:
                segi1, segi2 = int(r.partner1.residue.segment), int(r.partner2.residue.segment)
                pairs.setdefault(key1, []).append([segi1, segi2])
                cra_dict[key1+(segi1,)] = r.partner1
                cra_dict[key1+(segi2,)] = r.partner2

        if pairs: logger.write("Atoms on special position detected.")
        res_to_be_removed = []
        for key in sorted(pairs):
            logger.write(" Site: chain='{}' seq='{}{}' atom='{}' elem='{}' altloc='{}'".format(*key))
            # use graph to find connected components
            segs = sorted(set(sum(pairs[key], []))) # index->segid
            segd = dict([(s,i) for i,s in enumerate(segs)]) # reverse lookup
            g = numpy.zeros((len(segs),len(segs)), dtype=numpy.int)
            for p in pairs[key]:
                i, j = segd[p[0]], segd[p[1]]
                g[i,j] = g[j,i] = 1
            nc, labs = scipy.sparse.csgraph.connected_components(g, directed=False)
            groups = [[] for i in range(nc)] # list of segids
            for i, l in enumerate(labs): groups[l].append(segs[i])
            for group in groups:
                group.sort() # first segid will be kept
                sum_occ = sum([cra_dict[key+(i,)].atom.occ for i in group])
                logger.write("  multiplicity= {} occupancies_total= {:.2f} segids= {}".format(len(group), sum_occ, group))
                sum_pos = sum([cra_dict[key+(i,)].atom.pos for i in group], gemmi.Position(0,0,0))
                if len(group) < 2: continue # should never happen
                # modify first atom
                cra0 = cra_dict[key+(group[0],)]
                cra0.atom.occ = max(1, sum_occ)
                cra0.atom.pos = sum_pos/len(group)
                # remove remaining atoms
                for g in group[1:]:
                    cra = cra_dict[key+(g,)]
                    cra.residue.remove_atom(cra.atom.name, cra.atom.altloc, cra.atom.element)
                    if len(cra.residue) == 0: # empty residue needs to be removed
                        r_idx = [i for i, r in enumerate(cra.chain) if r==cra.residue]
                        res_to_be_removed.append((cra.chain, r_idx[0]))
                        
        chain_to_be_removed = []
        res_to_be_removed.sort(key=lambda x:([x[0].name, x[1]]))
        for chain, idx in reversed(res_to_be_removed):
            del chain[idx]
            if len(chain) == 0: # empty chain needs to be removed..
                c_idx = [i for i, c in enumerate(st[0]) if c==chain]
                chain_to_be_removed.append(c_idx[0])
        chain_to_be_removed.sort()
        for idx in reversed(chain_to_be_removed): #
            del st[0][idx] # we cannot use remove_chain() because ID may be duplicated

    # copy segment to subchain, as segment is not written to mmCIF file
    for chain in st[0]:
        for res in chain:
            res.subchain = res.segment
# expand_ncs()

def filter_helical_contacting(st, cutoff=5.):
    if len(st.ncs) == 0: return
    logger.write("Filtering out non-contacting helical copies with cutoff={:.2f} A".format(cutoff))
    st.setup_cell_images()
    ns = gemmi.NeighborSearch(st[0], st.cell, cutoff*2).populate()
    cs = gemmi.ContactSearch(cutoff)
    cs.ignore = gemmi.ContactSearch.Ignore.SameAsu
    results = cs.find_contacts(ns)
    indices = set([r.image_idx for r in results])
    logger.write(" contacting helical copies: {}".format(indices))
    ops = [st.ncs[i-1] for i in indices] # XXX is this correct? maybe yes as long as identity operator is not there
    st.ncs.clear()
    st.ncs.extend(ops)
# filter_helical_contacting()

def adp_analysis(st):
    logger.write("= ADP analysis =")
    all_B = []
    for i, mol in enumerate(st):
        logger.write("Model {}:".format(i))
        logger.write("            min    Q1   med    Q3   max")
        bs = []
        for chain in mol:
            for res in chain:
                for atom in res:
                    bs.append(atom.b_iso)
            qs = numpy.quantile(bs, [0,0.25,0.5,0.75,1])
            logger.write("Chain {:3s}".format(chain.name) + " {:5.1f} {:5.1f} {:5.1f} {:5.1f} {:5.1f}".format(*qs))
            all_B.extend(bs)

    qs = numpy.quantile(all_B, [0,0.25,0.5,0.75,1])
    logger.write("All       {:5.1f} {:5.1f} {:5.1f} {:5.1f} {:5.1f}".format(*qs))
    logger.write("")
# adp_analysis()        

def all_chain_ids(st):
    return [chain.name for model in st for chain in model]
# all_chain_ids()

def all_B(st):
    ret = []
    for mol in st:
        for cra in mol.all():
            ret.append(cra.atom.b_iso)

    return ret
# all_B()

def microheterogeneity_for_refmac(st, monlib):
    st.setup_entities()
    topo = gemmi.prepare_topology(st, monlib)
    mh_res = []
    chains = []
    icodes = {} # to avoid overlaps
    modifications = [] # return value
    
    # Check if microheterogeneity exists
    for chain in st[0]:
        for rg in chain.get_polymer().residue_groups():
            if len(rg) > 1:
                ress = [r for r in rg]
                chains.append(chain.name)
                mh_res.append(ress)
                ress_str = "/".join([str(r) for r in ress])
                logger.write("Microheterogeneity detected in chain {}: {}".format(chain.name, ress_str))

    if not mh_res: return []

    for chain in st[0]:
        for res in chain:
            if res.seqid.icode != " ":
                icodes.setdefault(chain.name, {}).setdefault(res.seqid.num, []).append(res.seqid.icode)
                
    def append_links(rinfo, prr, toappend):
        for pbond in filter(lambda f: (f.provenance==gemmi.Provenance.PrevLink and
                                       f.rkind==gemmi.RKind.Bond),
                            rinfo.forces):
            atoms = topo.bonds[pbond.index].atoms
            assert len(atoms) == 2
            found = None
            for i in range(2):
                if any(filter(lambda ra: atoms[i]==ra, prr)): found = i
            if found is not None:
                toappend.append([atoms[i], atoms[1-i]]) # prev atom, current atom
    # append_links()
    
    mh_res_all = sum(mh_res, [])
    mh_link = {}

    # Check links
    for cinfo in topo.chain_infos:
        for rinfo in cinfo.res_infos:
            # If this residue is microheterogeneous
            if rinfo.res in mh_res_all:
                for pr in rinfo.prev:
                    prr = pr.get(rinfo).res
                    mh_link.setdefault(rinfo.res, []).append([prr, "prev", pr.link, []])
                    append_links(rinfo, prr, mh_link[rinfo.res][-1][-1])
                    
            # Check if previous residue(s) is microheterogeneous
            for pr in rinfo.prev:
                prr = pr.get(rinfo).res
                if prr in mh_res_all:
                    mh_link.setdefault(prr, []).append([rinfo.res, "next", pr.link, []])
                    append_links(rinfo, prr, mh_link[prr][-1][-1])

    # Change IDs
    for chain_name, rr in zip(chains, mh_res):
        chars = string.ascii_uppercase
        # avoid already used inscodes
        if chain_name in icodes and rr[0].seqid.num in icodes[chain_name]:
            used_codes = set(icodes[chain_name][rr[0].seqid.num])
            chars = list(filter(lambda x: x not in used_codes, chars))
        for ir, r in enumerate(rr[1:]):
            modifications.append((chain_name, r.seqid.num, r.seqid.icode, chars[ir]))
            r.seqid.icode = chars[ir]

    # Update connections (LINKR)
    for chain_name, rr in zip(chains, mh_res):
        for r in rr:
            for p in mh_link.get(r, []):
                for atoms in p[-1]:
                    con = gemmi.Connection()
                    con.asu = gemmi.Asu.Same
                    con.type = gemmi.ConnectionType.Covale
                    con.link_id = p[2]
                    if p[1] == "prev":
                        p1 = gemmi.AtomAddress(chain_name, p[0].seqid, p[0].name, atoms[1].name, atoms[1].altloc)
                        p2 = gemmi.AtomAddress(chain_name, r.seqid, r.name, atoms[0].name, atoms[0].altloc)
                    else:
                        p1 = gemmi.AtomAddress(chain_name, r.seqid, r.name, atoms[1].name, atoms[1].altloc)
                        p2 = gemmi.AtomAddress(chain_name, p[0].seqid, p[0].name, atoms[0].name, atoms[0].altloc)
                        
                    con.partner1 = p1
                    con.partner2 = p2
                    logger.write("Adding link: {}".format(con))
                    st.connections.append(con)
        for r1, r2 in itertools.combinations(rr, 2):
            for a1 in set([a.altloc for a in r1]):
                for a2 in set([a.altloc for a in r2]):
                    con = gemmi.Connection()
                    con.asu = gemmi.Asu.Same
                    con.link_id = "gap"
                    # XXX altloc will be ignored when atom does not match.. grrr
                    con.partner1 = gemmi.AtomAddress(chain_name, r1.seqid, r1.name, "", a1)
                    con.partner2 = gemmi.AtomAddress(chain_name, r2.seqid, r2.name, "", a2)
                    st.connections.append(con)

    return modifications
# microheterogeneity_for_refmac()

def modify_inscodes_back(st, modifications):
    mods = dict([((cname,num,newcode), icode)for cname,num,icode,newcode in modifications])
    logger.write("Insertion codes to be modified back: {}".format(mods))
    for chain in st[0]:
        for res in chain:
            key = (chain.name, res.seqid.num, res.seqid.icode)
            if key in mods:
                logger.write(" Changing inscode '{}' to '{}' for {}/{}".format(res.seqid.icode,
                                                                           mods[key],
                                                                           chain.name,
                                                                           res.seqid.num))
                res.seqid.icode = mods[key]
# modify_inscodes_back()



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import argparse
import sys
import datetime
import pipes
import getpass
import platform
import servalcat.spa.sfcalc
import servalcat.spa.shiftback
import servalcat.spa.run_refmac
import servalcat.spa.fsc
import servalcat.spa.fofc
import servalcat.spa.shift_maps
import servalcat.utils.commands
#import servalcat.spa.bestmap

from servalcat.utils import logger

def main():
    
    parser = argparse.ArgumentParser(description="")
#Some useful commands:
#  trim        Trim maps and shift models to reduce file size
#  refina_spa  Refine CryoEM SPA structure using REFMAC5
#  fofc        Calculate updated map and Fo-Fc map using error estimates from half maps and model

    subparsers = parser.add_subparsers(dest="command")

    modules = dict(sfcalc=servalcat.spa.sfcalc,
                   shiftback=servalcat.spa.shiftback,
                   refine_spa=servalcat.spa.run_refmac,
                   fsc=servalcat.spa.fsc,
                   fofc=servalcat.spa.fofc,
                   trim=servalcat.spa.shift_maps,
                   #show=servalcat.utils.show,
                   util=servalcat.utils.commands,
                   #bestmap=servalcat.spa.bestmap,
                   )

    for n in modules:
        p = subparsers.add_parser(n)
        modules[n].add_arguments(p)

    args = parser.parse_args()
    
    if args.command == "util" and not args.subcommand:
        print("specify subcommand.")    
    elif args.command in modules:
        logger.set_file("servalcat.log")
        logger.write("# Servalcat ver. {}".format(servalcat.__version__))
        logger.write("# Started on {}".format(datetime.datetime.now()))
        logger.write("# Host: {} User: {}".format(platform.node(), getpass.getuser()))
        logger.write("# Command-line args:")
        logger.write("# {}".format(" ".join(map(lambda x: pipes.quote(x), sys.argv[1:]))))
        modules[args.command].main(args)
        logger.write("# Finished on {}\n".format(datetime.datetime.now()))
    else:
        parser.print_help()

# main()

if __name__ == "__main__":
    main()




"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
from servalcat.utils import logger
from servalcat.utils import model
import os
import subprocess
import gemmi
import numpy
import gzip

def splitext(path):
    if path.endswith((".bz2",".gz")):
        return os.path.splitext(path[:path.rindex(".")])
    else:
        return os.path.splitext(path)
# splitext()

def check_model_format(xyzin):
    # TODO check format actually
    # TODO mmjson is possible?
    ext = splitext(xyzin)[1]
    if ext.endswith("cif"):
        return ".mmcif"
    else:
        return ".pdb"
# check_model_format()

def write_mmcif(st, cif_out, cif_ref=None):
    """
    Refmac fails if _entry.id is longer than 80 chars including quotations
    """
    st_new = st.clone()
    print("Writing mmCIF file:", cif_out)
    if cif_ref:
        print("  using mmCIF metadata from:", cif_ref)
        groups = gemmi.MmcifOutputGroups(False)
        groups.ncs = True
        groups.atoms = True
        groups.cell = True
        groups.scale = True
        try:
            doc = read_cif_safe(cif_ref)
        except RuntimeError as e:
            # Sometimes refmac writes a broken mmcif file..
            logger.write("Error in mmCIF reading: {}".format(e))
            logger.write("  Give up using cif reference.")
            return write_mmcif(st, cif_out)
            
        blocks = list(filter(lambda b: b.find_loop("_atom_site.id"), doc))
        if len(blocks) == 0:
            logger.write("No _atom_site found in {}".format(e))
            logger.write("  Give up using cif reference.")
            return write_mmcif(st, cif_out)
        block = blocks[0]
        # to remove fract_transf_matrix. maybe we should keep some (like _atom_sites.solution_hydrogens)?
        # we do not want this because cell may be updated
        block.find_mmcif_category("_atom_sites.").erase()
        st_new.update_mmcif_block(block, groups)
        st_new.info["_entry.id"] = st_new.info["_entry.id"][:78]
        doc.write_file(cif_out)
    else:
        st_new.name = st_new.name[:78] # this will become _entry.id
        if "_entry.id" in st_new.info: st_new.info["_entry.id"] = st_new.info["_entry.id"][:78]
        st_new.make_mmcif_document().write_file(cif_out)
# write_mmcif()

def write_pdb(st, pdb_out):
    logger.write("Writing PDB file: {}".format(pdb_out))
    chain_id_lens = [len(x) for x in model.all_chain_ids(st)]
    if chain_id_lens and max(chain_id_lens) > 2:
        st = st.clone()
        st.shorten_chain_names()
    st.write_pdb(pdb_out, use_linkr=True)
# write_pdb()

def write_model(st, prefix=None, file_name=None, pdb=False, cif=False, cif_ref=None):
    if file_name:
        if file_name.endswith("cif"):
            write_mmcif(st, file_name, cif_ref)
        elif file_name.endswith(".pdb"):
            write_pdb(st, file_name)
        else:
            raise Exception("Cannot determine file format from file name: {}".format(file_name))
    else:            
        if cif:
            write_mmcif(st, prefix+".mmcif", cif_ref)
        if pdb:
            write_pdb(st, prefix+".pdb")
# write_model()

def read_shifts_txt(shifts_txt):
    ret = {}
    s = open(shifts_txt).read()
    s = s.replace("\n"," ").split()
    for i in range(len(s)-3):
        if s[i] in ("pdbin", "pdbout") and s[i+1] in ("cell", "shifts"):
            n = 6 if s[i+1] == "cell" else 3
            ret["{} {}".format(s[i], s[i+1])] = [float(x) for x in s[i+2:i+2+n]]

    return ret
# read_shifts_txt()

def read_ccp4_map(filename, setup=True, default_value=None, pixel_size=None):
    m = gemmi.read_ccp4_map(filename)
    g = m.grid
    grid_cell = [m.header_i32(x) for x in (8,9,10)]
    grid_start = [m.header_i32(x) for x in (5,6,7)]
    axis_pos = m.axis_positions()
    axis_letters = ["","",""]
    for i, l in zip(axis_pos, "XYZ"): axis_letters[i] = l
    spacings = [1./g.unit_cell.reciprocal().parameters[i]/grid_cell[i] for i in (0,1,2)]
    voxel_size = [g.unit_cell.parameters[i]/grid_cell[i] for i in (0,1,2)]
    label = m.header_str(57, 80)
    label = label[:label.find("\0")]
    logger.write("Reading CCP4/MRC map file {}".format(filename))
    logger.write("        Grid: {:4d} {:4d} {:4d}".format(*grid_cell))
    logger.write("    Map mode: {}".format(m.header_i32(4)))
    logger.write("       Start: {:4d} {:4d} {:4d}".format(*grid_start))
    logger.write("        Cell: {} {} {} {} {} {}".format(*g.unit_cell.parameters))
    logger.write("  Axis order: {}".format(" ".join(axis_letters)))
    logger.write(" Space group: {}".format(m.header_i32(23)))
    logger.write("     Spacing: {:.6f} {:.6f} {:.6f}".format(*spacings))
    logger.write("  Voxel size: {:.6f} {:.6f} {:.6f}".format(*voxel_size))
    logger.write("       Label: {}".format(label))
    logger.write("")

    if setup:
        if default_value is None: default_value = float("nan")
        m.setup(default_value)
        grid_start = [grid_start[i] for i in axis_pos]
        
    if pixel_size is not None:
        try:
            len(pixel_size)
        except TypeError:
            pixel_size = [pixel_size, pixel_size, pixel_size]
            
        logger.write("Overriding pixel size with {:.6f} {:.6f} {:.6f}".format(*pixel_size))
        orgc = m.grid.unit_cell.parameters
        new_abc = [orgc[i]*pixel_size[i]/voxel_size[i] for i in (0,1,2)]
        m.grid.unit_cell = gemmi.UnitCell(new_abc[0], new_abc[1], new_abc[2],
                                          orgc[3], orgc[4], orgc[5])
        logger.write(" New cell= {:.1f} {:.1f} {:.1f} {:.1f} {:.1f} {:.1f}".format(*m.grid.unit_cell.parameters))

    return [m.grid, grid_start] # should return original headers?
# read_ccp4_map()

def read_map_from_mtz(mtz_in, cols, grid_size=None, sample_rate=3):
    mtz = gemmi.read_mtz_file(mtz_in)
    d_min = mtz.resolution_high() # TODO get resolution for column?
    if grid_size is None:
        grid_size = mtz.get_size_for_hkl(sample_rate=sample_rate)
    F = mtz.get_f_phi_on_grid(cols[0], cols[1], grid_size)
    m = gemmi.transform_f_phi_grid_to_map(F)
    return d_min, m
# read_map_from_mtz()

def read_asu_data_from_mtz(mtz_in, cols):
    assert 0 < len(cols) < 3
    mtz = gemmi.read_mtz_file(mtz_in)
    sg = mtz.spacegroup
    miller = mtz.make_miller_array()
    f = mtz.column_with_label(cols[0])
    cell = mtz.get_cell(f.dataset_id)
    if len(cols) == 2:
        phi = mtz.column_with_label(cols[1])
        assert f.type == "F"
        assert phi.type == "P"
        phi = numpy.deg2rad(phi)
        f_comp = f * (numpy.cos(phi) + 1j * numpy.sin(phi))
        asu = gemmi.ComplexAsuData(cell, sg, miller, f_comp)
        return asu
    else:
        if f.is_integer():
            gr_t = gemmi.Int8AsuData # Why Int8?? XXX
        else:
            gr_t = gemmi.FloatAsuData
        
        asu = gr_t(cell, sg, miller, f)
        return asu
# read_asu_data_from_mtz()

def read_cif_safe(cif_in):
    ifs = gzip.open(cif_in) if cif_in.endswith(".gz") else open(cif_in)
    s = ifs.read()
    if "\0" in s: # Refmac occasionally writes \0 in some fileds..
        logger.write(" WARNING: null character detected. Replacing with '.'")
        s = s.replace("\0", ".")
    doc = gemmi.cif.read_string(s)
    return doc
# read_cif_safe()

def read_structure(xyz_in):
    spext = splitext(xyz_in)
    if spext[1].lower() in (".pdb", ".ent"):
        logger.write("Reading PDB file: {}".format(xyz_in))
        return gemmi.read_pdb(xyz_in)
    elif spext[1].lower() in (".cif", ".mmcif"):
        logger.write("Reading mmCIF file: {}".format(xyz_in))
        doc = read_cif_safe(xyz_in)
        blocks = list(filter(lambda b: b.find_loop("_atom_site.id"), doc))
        if len(blocks) == 0:
            raise RuntimeError("No block having _atom_site found")
        if len(blocks) > 1:
            logger.write(" WARNING: more than one block having _atom_site found. Will use first one.")
        return gemmi.make_structure_from_block(blocks[0])
    else:
        raise RuntimeError("Unsupported file type: {}".format(spext[1]))
# read_structure()

def read_structure_from_pdb_and_mmcif(xyz_in):
    st = read_structure(xyz_in)
    cif_ref = None
    spext = splitext(xyz_in)
    if spext[1] in (".pdb", ".ent"):
        cif_in = spext[0] + ".mmcif"
        if os.path.isfile(cif_in):
            print(" Will use mmcif metadata from {}".format(cif_in))
            cif_ref = cif_in
    elif spext[1] in (".cif", ".mmcif"):
        cif_ref = xyz_in
        pdb_in = spext[0] + ".pdb"
        if os.path.isfile(pdb_in):
            print(" Reading PDB REMARKS from {}".format(pdb_in))
            tmp = gemmi.read_structure(pdb_in)
            st.raw_remarks = tmp.raw_remarks

    if cif_ref is None and xyz_in.endswith("cif"):
        cif_ref = xyz_in
            
    return st, cif_ref
# read_structure_from_pdb_and_mmcif()

def merge_ligand_cif(cifs_in, cif_out):
    # TODO Check duplication?
    
    docs = [gemmi.cif.read(x) for x in cifs_in]
    tags = dict(comp=["_chem_comp.id"],
                link=["_chem_link.id"],
                mod=["_chem_mod.id"])
    found = dict(comp=0, link=0, mod=0)
    for d in docs:
        for k in tags:
            b = d.find_block("{}_list".format(k))
            if not b: continue
            found[k] += 1
            l = b.find_loop(tags[k][0]).get_loop()
            for t in l.tags:
                if t not in tags[k]: tags[k].append(t)
  
    doc = gemmi.cif.Document()
    for k in tags:
        if not found[k]: continue
        lst = doc.add_new_block("{}_list".format(k))
        loop = lst.init_loop("", tags[k])
        tags_for_find = [tags[k][0]] + ["?"+x for x in tags[k][1:]]
        
        for d in docs:
            b = d.find_block("{}_list".format(k))
            if not b: continue
            vals = b.find(tags_for_find)
            for v in vals:
                rl = [v.get(x) if v.has(x) else "." for x in range(len(tags[k]))]
                loop.add_row(rl)

    for d in docs:
        for b in d:
            if not b.name.endswith("_list"):
                doc.add_copied_block(b)

    doc.write_file(cif_out)
# merge_ligand_cif()



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import numpy
import pandas
import gemmi
from servalcat.utils import logger

dtypes64 = dict(i=numpy.int64, u=numpy.uint64, f=numpy.float64, c=numpy.complex128)
to64 = lambda x: x.astype(dtypes64.get(x.dtype.kind, x.dtype))

class Binner:
    def __init__(self, asu, style="relion"):
        if style == "relion":
            cell = asu.unit_cell
            max_cell_edge = numpy.max([cell.a, cell.b, cell.c])
            self.d_array = asu.make_d_array()
            self.bin_array = (max_cell_edge/self.d_array).astype(int)
            self.bins, self.bin_counts = numpy.unique(self.bin_array, return_counts=True)
        else:
            raise Exception("Non-supported binning type")
    # __init__()
# class Binner    

def df_from_asu_data(asu_data, label):
    df = pandas.DataFrame(data=asu_data.miller_array,
                          columns=["H","K","L"])
    if asu_data.value_array.dtype.names == ('value', 'sigma'):
        df[label] = to64(asu_data.value_array["value"])
        df["SIG"+label] = to64(asu_data.value_array["sigma"])
    else:
        df[label] = to64(asu_data.value_array)
    return df

def df_from_raw(miller_array, value_array, label):
    df = pandas.DataFrame(data=miller_array,
                          columns=["H","K","L"])
    df[label] = to64(value_array)
    return df

class HklData:
    def __init__(self, cell, sg, anomalous, df, binned_df=None):
        self.cell = cell
        self.sg = sg
        self.anomalous = anomalous
        self.df = df
        self.binned_df = binned_df
    # __init__()

    def update_cell(self, cell):
        # update d
        pass

    def switch_to_asu(self):
        # Need to care phases
        pass

    def copy(self):
        return HklData(self.cell, self.sg, self.anomalous,
                       self.df.copy(),
                       self.binned_df.copy() if self.binned_df is not None else None)

    def merge_asu_data(self, asu_data, label, common_only=True):
        if label in self.df:
            raise Exception("Duplicated label")
        
        df_tmp = df_from_asu_data(asu_data, label)
        if common_only:
            self.df = self.df.merge(df_tmp)
        else:
            self.df = self.df.merge(other.df, how="outer")
    # merge_asu_data()

    def miller_array(self):
        return self.df[["H","K","L"]]

    def calc_d(self):
        self.df["d"] = self.cell.calculate_d_array(self.miller_array())
    # calc_d()

    def calc_epsilon(self):
        self.df["epsilon"] = self.sg.operations().epsilon_factor_without_centering_array(self.miller_array())
    # calc_epsilon()

    def calc_centric(self):
        self.df["centric"] = self.sg.operations().centric_flag_array(self.miller_array()).astype(int)
    # calc_centric()
        
    def d_spacings(self):
        if "d" not in self.df or self.df.d.isnull().values.any():
            self.calc_d()
        return self.df.d
    # calc_d()

    def d_min_max(self):
        d = self.d_spacings()
        return min(d), max(d)
    # d_min_max()

    def setup_binning(self, n_bins, s_power=2):
        sp = self.d_spacings()**(-s_power)
        spmin, spmax = min(sp), max(sp)
        spstep = (spmax-spmin)/n_bins
        bin_limit_ds = numpy.arange(spmin, spmax, spstep)
        
        #bin_limit_ds = [sprange...] # left ends, and right end.
        if len(bin_limit_ds)==n_bins:
            bin_limit_ds = numpy.append(bin_limit_ds, bin_limit_ds[-1]+spstep)
        if bin_limit_ds[-1] < spmax:
            bin_limit_ds[-1] == spmax # difference should be very small..
        bin_limit_ds = bin_limit_ds**(-1/s_power)
        assert len(bin_limit_ds) == n_bins + 1

        bin_centers = [(bin_limit_ds[i]**(-s_power)+bin_limit_ds[i+1]**(-s_power))/2 for i in range(n_bins)]
        sprange = bin_limit_ds**(-s_power)

        self._bin_and_limits = []
        bin_number = numpy.zeros(len(sp), dtype=numpy.int)
        
        for i in range(1, len(sprange)):
            sel = numpy.where(numpy.logical_and(sprange[i-1]<=sp, sp <sprange[i]))[0]
            bin_number[sel] = i
            self._bin_and_limits.append((i, bin_limit_ds[i-1], bin_limit_ds[i]))

        self.df["bin"] = bin_number
        self.binned_df = pandas.DataFrame(data=list(range(max(self.df.bin)+1)),
                                          columns=["bin"])
    # setup_binning()
        
    def setup_relion_binning(self):
        max_edge = max(self.cell.parameters[:3])
        if "d" not in self.df or self.df.d.isnull().values.any():
            self.calc_d()
            
        self.df["bin"] = (max_edge/self.df.d).astype(numpy.int)
        self.binned_df = pandas.DataFrame(data=list(range(max(self.df.bin)+1)),
                                          columns=["bin"])
        self._bin_and_limits = []
        bin_numbers = set(self.df.bin)

        # Merge inner shells if too few
        while True:
            min_bin = min(bin_numbers)
            min_bin_sel = min_bin == self.df.bin
            n_min_bin = sum(min_bin_sel)
            if n_min_bin < 10:
                logger.write("Bin {} only has {} data. Merging with next bin.".format(min_bin, n_min_bin))
                self.df.loc[min_bin_sel, "bin"] = min_bin+1
                bin_numbers.remove(min_bin)
            else:
                break
                
        # TODO very slow
        for i_bin in bin_numbers:
            sel = self.df.bin == i_bin # selection may be kept, but df size may change later..?
            d_sel = self.df.d[sel]
            self._bin_and_limits.append((i_bin, max(d_sel), min(d_sel)))
    # setup_relion_binning()

    def bin_and_limits(self):
        return self._bin_and_limits
    # bin_and_limits()

    def binned_data_as_array(self, lab):
        vals = numpy.zeros(len(self.df.index), dtype=self.binned_df[lab].dtype)
        grouped = self.df.groupby("bin", sort=False)
        for b, g in grouped:
            vals[g.index] = self.binned_df[lab][b]
        return vals
    # binned_data_as_array()
    
    def merge(self, other, common_only=True):
        self.merge_df(other, common_only)
    # merge()

    def merge_df(self, other, common_only=True):
        # TODO check space group, cell, anomalous
        # TODO transform to asu with phase shifts
        # TODO check column labels. same names other than HKL?
        # 
        if common_only:
            self.df = self.df.merge(other)
        else:
            df = self.df.merge(other, indicator=True, how="outer")
            df_left = df[df._merge=="left_only"]
            df_right = df[df._merge=="right_only"]
            df_both = df[df._merge=="both"]
    # merge()

    def fft_map(self, label, grid_size=None, sample_rate=3):
        asu = gemmi.ComplexAsuData(self.cell,
                                   self.sg,
                                   self.miller_array(),
                                   self.df[label])
        if grid_size is None:
            grid_size = (0, 0, 0)

        ma = asu.transform_f_phi_to_map(sample_rate=sample_rate, exact_size=grid_size)
        return ma
    # fft_map()

    def d_eff(self, label):
        # Effective resolution definied using FSC
        fsc = self.binned_df[label]
        bin_counts = self.df.bin.value_counts()
        a = 0.
        for i_bin, bin_d_max, bin_d_min in self.bin_and_limits():
            a += bin_counts[i_bin] * fsc[i_bin]

        fac = (a/sum(bin_counts))**(1/3.)
        d_min = self.d_min_max()[0]
        ret = d_min/fac
        return ret
    # d_eff()



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
from servalcat.utils import logger
import os
import gemmi
import numpy

default_proton_scale = 1.13 # scale of X-proton distance to X-H(e) distance

def filename_in_monlib(monomer_dir, name):
    return os.path.join(monomer_dir, name[0].lower(), name+".cif")

def load_monomer_library(resnames, monomer_dir=None, cif_files=None):
    if monomer_dir is None:
        if "CLIBD_MON" not in os.environ:
            logger.write("ERROR: CLIBD_MON is not set")
        else:
            monomer_dir = os.environ["CLIBD_MON"]

    if cif_files is None:
        cif_files = []
        
    if not os.path.isdir(monomer_dir):
        logger.write("ERROR: not a directory: {}".format(monomer_dir))
        return

    if monomer_dir:
        resinlib = list(filter(lambda x: os.path.exists(filename_in_monlib(monomer_dir, x)), resnames))
        monlib = gemmi.read_monomer_lib(monomer_dir, resinlib)
    else:
        monlib = gemmi.MonLib()

    for f in cif_files: # TODO Support link!!
        logger.write("Reading monomer: {}".format(f))
        doc = gemmi.cif.read(f)
        for b in doc:
            if b.find_values("_chem_comp_atom.atom_id"):
                name = b.name.replace("comp_", "")
                if name in monlib.monomers:
                    logger.write("WARNING:: updating {} using {}".format(name, f))
                    del monlib.monomers[name]
                monlib.add_monomer_if_present(b)

    not_loaded = set(resnames).difference(monlib.monomers)
    if not_loaded:
        print("WARNING: monomers not loaded: {}".format(" ".join(not_loaded)))

    logger.write("Monomer library loaded: {} monomers, {} links, {} modifications".format(len(monlib.monomers),
                                                                                          len(monlib.links),
                                                                                          len(monlib.modifications)))
    logger.write("       Monomers: {}".format(" ".join([x for x in monlib.monomers])))
    logger.write("          Links: {}".format(" ".join([x for x in monlib.links])))
    logger.write("  Modifications: {}".format(" ".join([x for x in monlib.modifications])))
        
    return monlib
# load_monomer_library()

def check_monlib_support_nucleus_distances(monlib, resnames):
    good = True
    nucl_not_found = []
    for resn in resnames:
        if resn not in monlib.monomers:
            logger.write("ERROR: monomer information of {} not loaded".format(resn))
            good = False
        else:
            mon = monlib.monomers[resn]
            no_nuc = False
            for bond in mon.rt.bonds:
                is_h = (mon.get_atom(bond.id1.atom).is_hydrogen(), mon.get_atom(bond.id2.atom).is_hydrogen())
                if any(is_h) and bond.value_nucleus != bond.value_nucleus:
                    no_nuc = True
                    break
            if no_nuc:
                nucl_not_found.append(resn)
                good = False

    if nucl_not_found:
        logger.write("WARNING: nucleus distance is not found for: {}".format(" ".join(nucl_not_found)))
        logger.write("         default scale ({}) is used for nucleus distances.".format(default_proton_scale))
    return good
# check_monlib_support_nucleus_distances()

def add_hydrogens(st, monlib, pos="elec"):
    assert pos in ("elec", "nucl")
    
    st.setup_entities()
    topo = gemmi.prepare_topology(st, monlib, h_change=gemmi.HydrogenChange.ReAddButWater)
    if pos == "nucl":
        logger.write("Generating hydrogens at nucleus positions")
        restraints.check_monlib_support_nucleus_distances(monlib, resnames)
        topo.adjust_hydrogen_distances(gemmi.Restraints.DistanceOf.Nucleus, default_scale=default_proton_scale)
    else:
        logger.write("Generating hydrogens at electron positions")
# add_hydrogens()

def plane_deviations(atoms):
    pp = gemmi.find_best_plane(atoms)
    return [gemmi.get_distance_from_plane(a.pos, pp) for a in atoms]
# plane_deviations()

def show_all(st, monlib):
    st.setup_entities() # entity information is needed for links
    topo = gemmi.prepare_topology(st, monlib)
    
    lookup = dict([(x.atom, (x.chain, x.residue, x.atom)) for x in st[0].all()])
    label = lambda c, r, a: "{:4s}{:1s}{:3s}{:>2s}{:4d}{:1s}".format(a.name, a.altloc if a.altloc!="\x00" else " ",
                                                                     c.name,
                                                                     r.name, r.seqid.num, r.seqid.icode)

    topo_d = dict(bond=topo.bonds, angle=topo.angles, torsion=topo.torsions,)#, chir=topo.chirs)
    for k in topo_d:
        topo_k = topo_d[k]
        all_z = [b.calculate_z() for b in topo_k]
        all_d = [b.calculate() for b in topo_k]
        if k in ("angle", "torsion"): all_d = numpy.rad2deg(all_d)
        all_i = [b.restr.value for b in topo_k]
        perm = range(len(all_z))
        perm = numpy.argsort(all_z)
        if k == "torsion":
            tmp = [all_z[i] * topo_k[i].restr.esd for i in range(len(topo_k))]
            #print("TORSION_DEV=", tmp)
            rmsd = numpy.sqrt(numpy.average(numpy.array(tmp)**2))
        else:
            rmsd = numpy.sqrt(numpy.average((numpy.array(all_d)-all_i)**2))
        print("# {}_rmsd= {:.5f}".format(k, rmsd))
        for i in reversed(perm):
            t = topo_k[i]
            labs = " ".join([label(*lookup[x]) for x in t.atoms])
            b_ave = sum([x.b_iso for x in t.atoms])/len(t.atoms)
            print("{} {:.4f} {:.4f} {:.2f}".format(labs, all_d[i], all_z[i], b_ave))
            #if k=="torsion": print(t.restr.value, t.restr.esd, t.restr.period)
            #if all_z[i] < 10: break


    # Plane
    topo_k = topo.planes
    all_d = [numpy.sqrt(numpy.average(numpy.square(plane_deviations(t.atoms)))) for t in topo_k]
    perm = numpy.argsort(all_d)
    print("# planes_meandev= {:.4f}".format(numpy.average(all_d)))
    for i in reversed(perm):
        t = topo_k[i]
        labs = " ".join([label(*lookup[x]) for x in t.atoms])
        print("{} {:.4f}".format(labs, all_d[i]))

    # Chir
    topo_k = topo.chirs
    all_d = [b.calculate() for b in topo_k]
    all_i = [b.restr.sign for b in topo_k]
    all_c = [b.check() for b in topo_k]
    perm = numpy.argsort(all_c)
    print("# chirs")
    for i in reversed(perm):
        t = topo_k[i]
        labs = " ".join([label(*lookup[x]) for x in t.atoms])
        if not all_c[i]:
            print("{} {:.4f} {} {}".format(labs, all_d[i], all_i[i], all_c[i]))

if __name__ == "__main__":
    import sys
    model_in = sys.argv[1]
    st = gemmi.read_structure(model_in)
    monlib = load_monomer_library(st[0].get_all_residue_names(),
                                  cif_files=sys.argv[2:] if len(sys.argv)>2 else None)
    show_all(st, monlib)



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import subprocess
import numpy
import re
from servalcat.utils import logger
from servalcat.utils import fileio
from servalcat.utils import model
from servalcat.utils import generate_operators

def get_matrices_using_relion(sym):
    ps = subprocess.check_output(["relion_refine", "--sym", sym.strip(), "--print_symmetry_ops"])

    ret = []
    read_flag = -1
    for l in ps.splitlines():
        if b"R(" in l:
            ret.append(numpy.zeros((3,3)))
            read_flag = 0
        elif 0 <= read_flag < 3:
            ret[-1][read_flag,:] = [float(x) for x in l.split()]
            read_flag += 1
        elif read_flag >= 3:
            read_flag = -1

    return ret
# get_matrices_using_relion()

def operators_from_symbol(op):
    r = re.search("^((?P<a>I|T|O)|(?P<b>C|D)(?P<n>[0-9]+))$", op.upper())
    if not r:
        raise RuntimeError("Invalid point group symbol: {}".format(op))
    a, b, n = r.group("a"), r.group("b"), r.group("n")
    if n is not None and int(n) <= 0:
        raise RuntimeError("Non positive order given: {}".format(op))

    # RELION's conventions
    if a == "I":
        order1 = order2 = 5
        axis1 = numpy.array([0, 0.85065, 0.52573])
        axis2 = numpy.array([0.52573, 0, 0.85065])
    elif a == "O":
        order1 = order2 = 4
        axis1 = numpy.array([0,0,1.0])
        axis2 = numpy.array([0,1.0,0])
    elif a == "T":
        order1 = order2 = 3
        axis1 = numpy.array([0.0,0.0,1.0])
        axis2 = numpy.array([0.0,0.94280904,-0.33333333])
    elif b == "D":
        order1 = int(n)
        axis1 = numpy.array([0,0,1.0])
        order2 = 2
        axis2 = numpy.array([1.,0.,0.])
    elif b == "C":
        order1 = int(n)
        axis1 = numpy.array([0,0,1.0])
        order2 = 0
        axis2 = None
    return generate_operators.generate_all_elements(axis1, order1, axis2, order2)
# operators_from_symbol()

def show_operators_axis_angle(ops):
    for i, op in enumerate(ops):
        ax, ang = generate_operators.Rotation2AxisAngle_general(op)
        logger.write(" operator {:3d} angle= {:7.3f} deg axis= {}".format(i+1, numpy.rad2deg(ang), list(ax)))
# show_operators_axis_angle()

def read_helical_parameters_from_mmcif(cif_in):
    doc = fileio.read_cif_safe(cif_in)
    b = list(filter(lambda b: b.find_loop("_atom_site.id"), doc))[0]
    deltaphi = b.find_value("_em_helical_entity.angular_rotation_per_subunit")
    deltaz = b.find_value("_em_helical_entity.axial_rise_per_subunit")
    deltaphi, deltaz = float(deltaphi), float(deltaz)
    axsym = b.find_value("_em_helical_entity.axial_symmetry")
    return axsym, deltaphi, deltaz
# read_helical_parameters_from_mmcif()

def generate_helical_operators(st, start_xyz, center, axsym, deltaphi, deltaz, padding=1):
    if not axsym: axsym = "C1"
    _, _, axtrs = operators_from_symbol(axsym)
    all_z = [cra.atom.pos.z for cra in st[0].all()]
    min_z, max_z = min(all_z), max(all_z)
    min_n, max_n = int((min_z-padding-start_xyz[2])/deltaz), int((st.cell.c+start_xyz[2]-max_z-padding)/deltaz)
    ops = []
    for i in range(-min_n, max_n+1):
        deg = deltaphi*i
        t = numpy.deg2rad(deg)
        m = generate_operators.AngleAxis2rotatin(numpy.array([0,0,1.]), t)
        s = numpy.array([0,0,deltaz*i])
        for a in axtrs:
            mat = numpy.dot(m, a)
            news = s + numpy.dot(mat, -center) + center
            tr = gemmi.Transform(gemmi.Mat33(mat), gemmi.Vec3(*news))
            newop = gemmi.NcsOp(tr, str(len(ops)+1), tr.is_identity())
            ops.append(newop)

    return ops
# generate_helical_operators()

"""
def write_ncsc_for_refmac(file_name, matrices, xyz_in=None, map_in=None):
    if xyz_in:
        st = gemmi.read_structure(xyz_in)
        cell = st.cell
    if map_in:
        ma = gemmi.read_ccp4_map(map_in)
        cell = ma.grid.unit_cell
        
    A = numpy.array(cell.orthogonalization_matrix.tolist())
    center = numpy.sum(A,axis=1) / 2
    
    ofs = open(file_name, "w")
    for m in matrices:
        transl = numpy.dot(m, -center) + center
        m_str = " ".join([str(x) for x in m.flatten()])
        t_str = " ".join([str(x) for x in transl])
        ofs.write("ncsc matrix {} {}\n".format(m_str, t_str))
    ofs.close()
# write_ncsc_for_refmac()
"""

def make_NcsOps_from_matrices(matrices, cell=None, center=None):
    if center is None:
        A = numpy.array(cell.orthogonalization_matrix.tolist())
        center = numpy.sum(A,axis=1) / 2

    center = gemmi.Vec3(*center)
    ops = []
    for i, m in enumerate(matrices):
        m = gemmi.Mat33(m)
        transl = m.multiply(-center) + center
        op = gemmi.NcsOp(gemmi.Transform(m, transl), str(i+1), m.is_identity())
        ops.append(op)
        
    return ops
# make_NcsOps_from_matrices()

def write_NcsOps_for_refmac(ncs_ops, file_name):
    def make_line(tr):
        m = tr.mat.tolist()
        m_str = " ".join([str(m[i][j]) for i in range(3) for j in range(3)])
        t_str = " ".join([str(x) for x in tr.vec.tolist()])
        return "ncsc matrix {} {}\n".format(m_str, t_str)
    
    ofs = open(file_name, "w")

    # REFMAC requires identity op
    if not any([x.tr.is_identity() for x in ncs_ops]):
        ofs.write(make_line(gemmi.Transform()))

    for op in ncs_ops: ofs.write(make_line(op.tr))
    ofs.close()
# write_NcsOps_for_refmac()

# TODO def euler2matrix(euler):
# TODO def polar2matrix(polar):

def parse_ncsc_keywords(kwd_str):
    # FIXME handle lines ending with -
    lines = kwd_str.splitlines()
    ret = []
    for l in lines:
        l = l.split()
        if len(l) < 3: continue
        if l[0].lower().startswith("ncsc"):
            if l[1].lower().startswith("matr"):
                vals = [float(x) for x in l[2:]]
                if len(vals) != 12:
                    print("Bad nsc matrix line: {}".format(" ".join(l)))
                    continue
                op = gemmi.NcsOp()
                op.tr.mat.fromlist([vals[3*x:3*x+3] for x in range(3)])
                op.tr.vec.fromlist(vals[9:])
                op.id = str(len(ret)+1)
                op.given = op.tr.is_identity()
                ret.append(op)
            elif l[1].lower().startswith("eule"):
                pass # TODO
            elif l[1].lower().startswith("pola"):
                pass # TODO
    return ret
# parse_ncsc_keywords()

def apply_shift_for_ncsops(ncsops, shift):
    new_ops = []
    s = gemmi.Vec3(*shift)
    for op in ncsops:
        newt = op.tr.vec + s - op.tr.mat.multiply(s)
        newop = gemmi.NcsOp(gemmi.Transform(op.tr.mat, newt), op.id, op.given)
        new_ops.append(newop)

    return new_ops
# apply_shift_for_ncsops()

def write_symmetry_expanded_model(st, prefix, pdb=False, cif=False, cif_ref=None):
    st_new = st.clone()
    model.expand_ncs(st_new)
    fileio.write_model(st_new, prefix=prefix, pdb=pdb, cif=cif, cif_ref=cif_ref)
# write_symmetry_expanded_model()
